"""Connection tests."""

import asyncio
from unittest.mock import AsyncMock, Mock, call, patch

import pytest

from pypck import inputs
from pypck.connection import (
    PchkAuthenticationError,
    PchkConnectionFailedError,
    PchkConnectionManager,
    PchkConnectionRefusedError,
    PchkLicenseError,
)
from pypck.lcn_addr import LcnAddr
from pypck.lcn_defs import AcknowledgeErrorCode, LcnEvent
from pypck.pck_commands import PckGenerator

from .conftest import HOST, PASSWORD, PORT, USERNAME, MockPchkConnectionManager


async def test_close_without_connect(pypck_client: MockPchkConnectionManager) -> None:
    """Test closing of PchkConnectionManager without connecting."""
    await pypck_client.async_close()


@patch.object(PchkConnectionManager, "open_connection")
@patch.object(PchkConnectionManager, "scan_segment_couplers")
async def test_async_connect(
    mock_scan_segment_couplers: AsyncMock,
    mock_open_connection: AsyncMock,
) -> None:
    """Test successful connection."""
    pypck_client = PchkConnectionManager(HOST, PORT, USERNAME, PASSWORD)
    connect_task = asyncio.create_task(pypck_client.async_connect())
    await asyncio.sleep(0)

    pypck_client.license_error_future.set_result(True)
    pypck_client.authentication_completed_future.set_result(True)
    pypck_client.segment_scan_completed_event.set()

    await connect_task

    mock_scan_segment_couplers.assert_awaited()
    mock_open_connection.assert_awaited()
    assert pypck_client.is_ready()


@patch.object(PchkConnectionManager, "ping")
@patch.object(PchkConnectionManager, "open_connection")
@patch.object(PchkConnectionManager, "scan_segment_couplers")
@patch.object(PchkConnectionManager, "send_command")
async def test_successful_connection_procedure(
    mock_send_command: AsyncMock,
    mock_scan_segment_couplers: AsyncMock,
    mock_open_connection: AsyncMock,
    mock_ping: AsyncMock,
) -> None:
    """Test successful connection procedure."""
    pypck_client = PchkConnectionManager(HOST, PORT, USERNAME, PASSWORD)
    connect_task = asyncio.create_task(pypck_client.async_connect())
    await asyncio.sleep(0)

    await pypck_client.async_process_input(inputs.AuthUsername())
    mock_send_command.assert_awaited_with(USERNAME, to_host=True)

    await pypck_client.async_process_input(inputs.AuthPassword())
    mock_send_command.assert_awaited_with(PASSWORD, to_host=True)

    await pypck_client.async_process_input(inputs.AuthOk())
    mock_send_command.assert_awaited_with(PckGenerator.set_dec_mode(), to_host=True)
    assert pypck_client.authentication_completed_future.result()

    await pypck_client.async_process_input(inputs.DecModeSet())
    mock_send_command.assert_awaited_with(
        PckGenerator.set_operation_mode(
            pypck_client.dim_mode, pypck_client.status_mode
        ),
        to_host=True,
    )
    assert pypck_client.license_error_future.result()

    await connect_task

    mock_open_connection.assert_awaited()
    mock_scan_segment_couplers.assert_awaited()
    mock_ping.assert_awaited()


@pytest.mark.parametrize("side_effect", [ConnectionRefusedError, OSError])
async def test_connection_error(side_effect: ConnectionRefusedError | OSError) -> None:
    """Test connection error."""
    with (
        patch.object(PchkConnectionManager, "open_connection", side_effect=side_effect),
        pytest.raises(PchkConnectionRefusedError),
    ):
        pypck_client = PchkConnectionManager(HOST, PORT, USERNAME, PASSWORD)
        await pypck_client.async_connect()


@patch.object(PchkConnectionManager, "open_connection")
async def test_authentication_error(mock_open_connection: AsyncMock) -> None:
    """Test wrong login credentials."""
    pypck_client = PchkConnectionManager(HOST, PORT, USERNAME, PASSWORD)
    connect_task = asyncio.create_task(pypck_client.async_connect())
    await asyncio.sleep(0)
    await pypck_client.async_process_input(inputs.AuthFailed())

    with (
        pytest.raises(PchkAuthenticationError),
    ):
        await connect_task


@patch.object(PchkConnectionManager, "open_connection")
async def test_license_error(mock_open_connection: AsyncMock) -> None:
    """Test wrong login credentials."""
    pypck_client = PchkConnectionManager(HOST, PORT, USERNAME, PASSWORD)
    connect_task = asyncio.create_task(pypck_client.async_connect())
    await asyncio.sleep(0)
    await pypck_client.async_process_input(inputs.LicenseError())

    with (
        pytest.raises(PchkLicenseError),
    ):
        await connect_task


@patch.object(PchkConnectionManager, "open_connection")
async def test_timeout_error(mock_open_connection: AsyncMock) -> None:
    """Test timeout when connecting."""
    with pytest.raises(PchkConnectionFailedError):
        pypck_client = PchkConnectionManager(HOST, PORT, USERNAME, PASSWORD)
        await pypck_client.async_connect(timeout=0)


async def test_lcn_connected(pypck_client: MockPchkConnectionManager) -> None:
    """Test lcn connected events."""
    event_callback = Mock()
    pypck_client.register_for_events(event_callback)
    await pypck_client.async_connect()

    # bus disconnected
    await pypck_client.async_process_input(inputs.LcnConnState(is_lcn_connected=False))
    assert not pypck_client.is_lcn_connected
    event_callback.assert_has_calls(
        (call(LcnEvent.BUS_CONNECTION_STATUS_CHANGED), call(LcnEvent.BUS_DISCONNECTED))
    )

    # bus connected
    await pypck_client.async_process_input(inputs.LcnConnState(is_lcn_connected=True))
    assert pypck_client.is_lcn_connected
    event_callback.assert_has_calls(
        (call(LcnEvent.BUS_CONNECTION_STATUS_CHANGED), call(LcnEvent.BUS_CONNECTED))
    )


async def test_new_module_on_input(
    pypck_client: MockPchkConnectionManager,
) -> None:
    """Test new module detection on serial input."""
    await pypck_client.async_connect()
    address = LcnAddr(0, 7, False)
    assert address not in pypck_client.device_connections

    await pypck_client.async_process_input(
        inputs.ModAck(address, AcknowledgeErrorCode.OK)
    )

    assert address in pypck_client.device_connections



"""Classes to store module and group addresses."""

from dataclasses import dataclass


@dataclass(frozen=True)
class LcnAddr:
    """Represents a LCN address (module or group).

    If the segment id is 0, the address object points to modules/groups which
    are in the segment where the bus coupler is connected to. This is also the
    case if no segment coupler is present at all.

    :param    int    seg_id:    Segment id
                                (0 = Local,
                                1..2 = Not allowed (but "seen in the wild")
                                3 = Broadcast,
                                4 = Status messages,
                                5..127,
                                128 = Segment-bus disabled (valid value))
    :param    bool   is_group:  Indicates whether address point to a module
                                (False) or a group (True)


    If address represents a **module**:

    :param    int    addr_id:   Module id
                                (1 = LCN-PRO,
                                2 = LCN-GVS/LCN-W,
                                4 = PCHK,
                                5..254,
                                255 = Unprog. (valid, but irrelevant here))


    If address represents a **group**:

    :param    int    addr_id:   Group id
                                (3 = Broadcast,
                                4 = Status messages,
                                5..254)
    """

    seg_id: int
    addr_id: int
    is_group: bool = False

    def get_physical_seg_id(self, local_seg_id: int) -> int:
        """Get the physical segment id ("local" segment replaced with 0).

        Can be used to send data into the LCN bus.

        :param    int    local_seg_id:    The segment id of the local segment

        :return:    The physical segment id
        :rtype:     int
        """
        return 0 if (self.seg_id == local_seg_id) else self.seg_id

    @classmethod
    def from_string(cls, addr_str: str) -> "LcnAddr":
        """Parse an LCN address from a string.

        :param    str addr_str:    The address string (e.g. "m001002" or "g003004")

        :return:    The parsed address
        :rtype:     LcnAddr
        """
        if len(addr_str) != 7:
            raise ValueError(f"Invalid address string: {addr_str}")
        kind = addr_str[0].lower()
        if kind not in ("m", "g"):
            raise ValueError(f"Invalid address kind: {kind}")
        is_group = kind == "g"
        try:
            seg_id = int(addr_str[1:4])
            addr_id = int(addr_str[4:7])
        except ValueError as err:
            raise ValueError(f"Invalid address string: {addr_str}") from err
        return cls(seg_id=seg_id, addr_id=addr_id, is_group=is_group)

    def to_string(self) -> str:
        """Format the address as a string."""
        kind = "g" if self.is_group else "m"
        return f"{kind}{self.seg_id:03d}{self.addr_id:03d}"

    def __str__(self) -> str:
        """Format the address as a string."""
        return self.to_string()



"""Core testing functionality."""

import asyncio
from typing import Any, cast
from unittest.mock import AsyncMock, patch

import pytest

import pypck
from pypck.connection import PchkConnectionManager
from pypck.device import DeviceConnection
from pypck.lcn_addr import LcnAddr
from pypck.pck_commands import PckGenerator

HOST = "127.0.0.1"
PORT = 4114
USERNAME = "lcn"
PASSWORD = "lcn"


async def wait_until_called(
    mock: AsyncMock,
    *expected_args: Any,
    timeout: float = 1.0,
    **expected_kwargs: Any,
) -> None:
    """Wait that AsyncMock gets called with given arguments."""
    event = asyncio.Event()

    async def side_effect(*args: Any, **kwargs: Any) -> None:
        """Set the event when the mock is called."""
        if (len(expected_args) == 0 or args == expected_args) and (
            len(expected_kwargs) == 0 or kwargs == expected_kwargs
        ):
            event.set()

    mock.side_effect = side_effect

    await asyncio.wait_for(event.wait(), timeout=timeout)


class MockDeviceConnection(DeviceConnection):
    """Fake a LCN device connection."""

    send_command = AsyncMock(return_value=True)


class MockPchkConnectionManager(PchkConnectionManager):
    """Fake connection handler."""

    is_lcn_connected: Any

    async def async_connect(self, timeout: float = 30) -> None:
        """Mock establishing a connection to PCHK."""
        self.authentication_completed_future.set_result(True)
        self.license_error_future.set_result(True)
        self.segment_scan_completed_event.set()

    async def async_close(self) -> None:
        """Mock closing a connection to PCHK."""

    @patch.object(pypck.connection, "DeviceConnection", MockDeviceConnection)
    def get_device_connection(self, addr: LcnAddr) -> DeviceConnection:
        """Get LCN device connection."""
        return super().get_device_connection(addr)

    scan_modules = AsyncMock()
    send_command = AsyncMock()


def encode_pck(pck: str) -> bytes:
    """Encode the given PCK string as PCK binary string."""
    return (pck + PckGenerator.TERMINATION).encode()


@pytest.fixture
async def pypck_client() -> MockPchkConnectionManager:
    """Create a mock PCHK connection manager."""
    return MockPchkConnectionManager(HOST, PORT, USERNAME, PASSWORD)


@pytest.fixture
async def module10(
    pypck_client: MockPchkConnectionManager,
) -> MockDeviceConnection:
    """Create test module with addr_id 10."""
    lcn_addr = LcnAddr(0, 10, False)
    with patch.object(
        MockDeviceConnection, "_request_device_properties"
    ) as mock_request_device_properties:
        module = cast(
            MockDeviceConnection, pypck_client.get_device_connection(lcn_addr)
        )
        await wait_until_called(cast(AsyncMock, mock_request_device_properties))

    module.send_command.reset_mock()
    return module



"""Tests for module."""

import asyncio
from itertools import chain

import pytest

from pypck import inputs, lcn_defs
from pypck.device import Serials
from pypck.lcn_addr import LcnAddr
from pypck.pck_commands import PckGenerator

from .conftest import MockDeviceConnection, wait_until_called

RELAY_STATES = [True, False, True, False, True, False, True, False]
BINARY_SENSOR_STATES = [True, False, True, False, True, False, True, False]
LED_STATES = [
    lcn_defs.LedStatus.ON,
    lcn_defs.LedStatus.OFF,
    lcn_defs.LedStatus.BLINK,
    lcn_defs.LedStatus.FLICKER,
    lcn_defs.LedStatus.ON,
    lcn_defs.LedStatus.OFF,
    lcn_defs.LedStatus.BLINK,
    lcn_defs.LedStatus.FLICKER,
    lcn_defs.LedStatus.ON,
    lcn_defs.LedStatus.OFF,
    lcn_defs.LedStatus.BLINK,
    lcn_defs.LedStatus.FLICKER,
]
LOGIC_OPS_STATES = [
    lcn_defs.LogicOpStatus.ALL,
    lcn_defs.LogicOpStatus.NONE,
    lcn_defs.LogicOpStatus.SOME,
    lcn_defs.LogicOpStatus.NONE,
]
LOCKED_KEY_STATES = [
    [True, False, True, False, True, False, True, False],
    [False, True, False, True, False, True, False, True],
    [True, True, True, True, False, False, False, False],
    [False, False, False, False, True, True, True, True],
]

#
# Status requests
#


@pytest.mark.parametrize(
    "output_port",
    [
        lcn_defs.OutputPort.OUTPUT1,
        lcn_defs.OutputPort.OUTPUT2,
        lcn_defs.OutputPort.OUTPUT3,
        lcn_defs.OutputPort.OUTPUT4,
    ],
)
async def test_request_status_output(
    module10: MockDeviceConnection, output_port: lcn_defs.OutputPort
) -> None:
    """Test requesting the output status of a module."""
    request_task = asyncio.create_task(module10.request_status_output(output_port))

    await wait_until_called(module10.send_command)
    await module10.async_process_input(
        inputs.ModStatusOutput(module10.addr, output_port.value, 50.0)
    )

    result = await request_task

    assert isinstance(result, inputs.ModStatusOutput)
    assert result.physical_source_addr == module10.addr
    assert result.output_id == output_port.value
    assert result.percent == 50.0


async def test_request_status_relays(module10: MockDeviceConnection) -> None:
    """Test requesting the relays status of a module."""
    request_task = asyncio.create_task(module10.request_status_relays())

    await wait_until_called(module10.send_command)
    await module10.async_process_input(
        inputs.ModStatusRelays(module10.addr, RELAY_STATES)
    )

    result = await request_task

    assert isinstance(result, inputs.ModStatusRelays)
    assert result.physical_source_addr == module10.addr
    assert result.states == RELAY_STATES


@pytest.mark.parametrize(
    "motor",
    [
        lcn_defs.MotorPort.MOTOR1,
        lcn_defs.MotorPort.MOTOR2,
        lcn_defs.MotorPort.MOTOR3,
        lcn_defs.MotorPort.MOTOR4,
    ],
)
async def test_request_status_motor_position(
    module10: MockDeviceConnection, motor: lcn_defs.MotorPort
) -> None:
    """Test requesting the motors status of a module."""
    request_task = asyncio.create_task(
        module10.request_status_motor_position(motor, lcn_defs.MotorPositioningMode.BS4)
    )

    await wait_until_called(module10.send_command)
    await module10.async_process_input(
        inputs.ModStatusMotorPositionBS4(module10.addr, motor.value, 50.0)
    )

    result = await request_task

    assert isinstance(result, inputs.ModStatusMotorPositionBS4)
    assert result.physical_source_addr == module10.addr
    assert result.motor == motor.value
    assert result.position == 50.0


async def test_request_status_binary_sensors(module10: MockDeviceConnection) -> None:
    """Test requesting the binary sensors status of a module."""
    request_task = asyncio.create_task(module10.request_status_binary_sensors())

    await wait_until_called(module10.send_command)
    await module10.async_process_input(
        inputs.ModStatusBinSensors(module10.addr, BINARY_SENSOR_STATES)
    )

    result = await request_task

    assert isinstance(result, inputs.ModStatusBinSensors)
    assert result.physical_source_addr == module10.addr
    assert result.states == BINARY_SENSOR_STATES


@pytest.mark.parametrize(
    "variable, software_serial",
    [
        *[
            (variable, 0x170206)
            for variable in lcn_defs.Var.variables_new()
            + lcn_defs.Var.set_points()
            + list(chain(*lcn_defs.Var.thresholds_new()))
            + lcn_defs.Var.s0s()
        ],
        *[
            (variable, 0x170000)
            for variable in lcn_defs.Var.variables_old()
            + lcn_defs.Var.set_points()
            + list(chain(*lcn_defs.Var.thresholds_old()))
        ],
    ],
)
async def test_request_status_variable(
    module10: MockDeviceConnection, variable: lcn_defs.Var, software_serial: int
) -> None:
    """Test requesting the variable status of a module."""
    module10.serials.software_serial = software_serial
    request_task = asyncio.create_task(module10.request_status_variable(variable))

    await wait_until_called(module10.send_command)
    response_variable = (
        variable
        if lcn_defs.Var.has_type_in_response(variable, software_serial)
        else lcn_defs.Var.UNKNOWN
    )
    await module10.async_process_input(
        inputs.ModStatusVar(
            module10.addr, response_variable, lcn_defs.VarValue.from_native(50)
        )
    )

    result = await request_task

    assert isinstance(result, inputs.ModStatusVar)
    assert result.physical_source_addr == module10.addr
    assert result.var == variable
    assert result.value == lcn_defs.VarValue.from_native(50)


async def test_request_status_leds_and_logic_ops(
    module10: MockDeviceConnection,
) -> None:
    """Test requesting the LED and logic operations status of a module."""
    request_task = asyncio.create_task(module10.request_status_leds_and_logic_ops())

    await wait_until_called(module10.send_command)
    await module10.async_process_input(
        inputs.ModStatusLedsAndLogicOps(module10.addr, LED_STATES, LOGIC_OPS_STATES)
    )

    result = await request_task

    assert isinstance(result, inputs.ModStatusLedsAndLogicOps)
    assert result.physical_source_addr == module10.addr
    assert result.states_led == LED_STATES
    assert result.states_logic_ops == LOGIC_OPS_STATES


async def test_request_status_locked_keys(module10: MockDeviceConnection) -> None:
    """Test requesting the locker keys status of a module."""
    request_task = asyncio.create_task(module10.request_status_locked_keys())

    await wait_until_called(module10.send_command)
    await module10.async_process_input(
        inputs.ModStatusKeyLocks(module10.addr, LOCKED_KEY_STATES)
    )

    result = await request_task

    assert isinstance(result, inputs.ModStatusKeyLocks)
    assert result.physical_source_addr == module10.addr
    assert result.states == LOCKED_KEY_STATES


async def test_request_serials(module10: MockDeviceConnection) -> None:
    """Test requesting serials of a module."""
    request_task = asyncio.create_task(module10.request_serials())

    await wait_until_called(module10.send_command, False, PckGenerator.request_serial())
    await module10.async_process_input(
        inputs.ModSn(
            module10.addr,
            hardware_serial=0x1A20A1234,
            manu=0x1,
            software_serial=0x190B11,
            hardware_type=lcn_defs.HardwareType.SH_PLUS,
        )
    )
    result = await request_task

    assert isinstance(result, Serials)
    assert result.hardware_serial == 0x1A20A1234
    assert result.manu == 0x1
    assert result.software_serial == 0x190B11
    assert result.hardware_type == lcn_defs.HardwareType.SH_PLUS


@pytest.mark.parametrize(
    "command, blocks, text",
    [
        ("N", ["1234567890", "ABCDEFGHIJ"], "1234567890ABCDEFGHIJ"),
        ("N", ["12345678", "ABCDEFGH"], "12345678  ABCDEFGH"),
        ("N", ["1234567890", ""], "1234567890"),
        (
            "K",
            ["123456789012", "ABCDEFGHIJKL", "abcdefghijkl"],
            "123456789012ABCDEFGHIJKLabcdefghijkl",
        ),
        (
            "K",
            ["1234567890", "ABCDEFGHIJ", "abcdefghij"],
            "1234567890  ABCDEFGHIJ  abcdefghij",
        ),
        ("K", ["1234567890", "", ""], "1234567890"),
        (
            "O",
            ["123456789012", "ABCDEFGHIJKL", "abcdefghijkl", "123456789012"],
            "123456789012ABCDEFGHIJKLabcdefghijkl123456789012",
        ),
        (
            "O",
            ["1234567890", "ABCDEFGHIJ", "abcdefghij", "1234567890"],
            "1234567890  ABCDEFGHIJ  abcdefghij  1234567890",
        ),
        ("O", ["1234567890", "", "", ""], "1234567890"),
    ],
)
async def test_request_text(
    command: str, blocks: list[str], text: str, module10: MockDeviceConnection
) -> None:
    """Test requesting the name, comment or oem_text of a module."""
    match command:
        case "N":
            request_task = asyncio.create_task(module10.request_name())
        case "K":
            request_task = asyncio.create_task(module10.request_comment())
        case "O":
            request_task = asyncio.create_task(module10.request_oem_text())

    for idx, block in enumerate(blocks):
        await wait_until_called(module10.send_command)
        await module10.async_process_input(
            inputs.ModNameComment(
                module10.addr,
                command=command,
                block_id=idx,
                text=block,
            )
        )

    result = await request_task

    assert isinstance(result, str)
    assert result == text


@pytest.mark.parametrize("dynamic", [True, False])
async def test_request_group_memberships(
    dynamic: bool, module10: MockDeviceConnection
) -> None:
    """Test requesting group memberships of a module."""
    addresses = [LcnAddr(0, 7 + id, False) for id in range(3)]
    request_task = asyncio.create_task(module10.request_group_memberships(dynamic))

    await wait_until_called(module10.send_command)
    await module10.async_process_input(
        inputs.ModStatusGroups(module10.addr, dynamic, 12, addresses)
    )
    result = await request_task

    assert isinstance(result, set)
    assert result == set(addresses)



"""Tests for command generation directed at bus modules and groups."""

from typing import Any

import pytest

from pypck.lcn_addr import LcnAddr
from pypck.lcn_defs import (
    BeepSound,
    KeyLockStateModifier,
    LedStatus,
    MotorPositioningMode,
    MotorReverseTime,
    MotorStateModifier,
    OutputPort,
    OutputPortDimMode,
    OutputPortStatusMode,
    RelayPort,
    RelayStateModifier,
    RelVarRef,
    SendKeyCommand,
    ThresholdLockStateModifier,
    TimeUnit,
    Var,
)
from pypck.pck_commands import PckGenerator

NEW_VAR_SW_AGE = 0x170206

COMMANDS: dict[str | bytes, Any] = {
    # Host commands
    **{
        f"^ping{counter:d}": (PckGenerator.ping, counter)
        for counter in (1, 10, 100, 1000, 10000)
    },
    "!CHD": (PckGenerator.set_dec_mode,),
    **{
        f"!OM{dim_mode.value:d}{status_mode.value:s}": (
            PckGenerator.set_operation_mode,
            dim_mode,
            status_mode,
        )
        for dim_mode in OutputPortDimMode
        for status_mode in OutputPortStatusMode
    },
    # Command address header
    **{
        f">{addr_type:s}{seg_id:03d}{addr_id:03d}{separator:s}": (
            PckGenerator.generate_address_header,
            LcnAddr(seg_id, addr_id, addr_type == "G"),
            0,
            separator == "!",
        )
        for seg_id in (0, 5, 10, 100)
        for addr_id in (5, 10, 100)
        for addr_type in ("G", "M")
        for separator in ("!", ".")
    },
    ">M000021.": (
        PckGenerator.generate_address_header,
        LcnAddr(7, 21, False),
        7,
        False,
    ),
    # Other module commands
    "LEER": (PckGenerator.empty,),
    **{
        f"PIN{count:03d}": (PckGenerator.beep, BeepSound.NORMAL, count)
        for count in range(1, 16)
    },
    **{
        f"PIS{count:03d}": (PckGenerator.beep, BeepSound.SPECIAL, count)
        for count in range(1, 16)
    },
    # General status commands
    "SK": (PckGenerator.segment_coupler_scan,),
    "SN": (PckGenerator.request_serial,),
    **{f"NMN{block + 1}": (PckGenerator.request_name, block) for block in range(2)},
    **{f"NMK{block + 1}": (PckGenerator.request_comment, block) for block in range(3)},
    **{f"NMO{block + 1}": (PckGenerator.request_oem_text, block) for block in range(4)},
    "GP": (PckGenerator.request_group_membership_static,),
    "GD": (PckGenerator.request_group_membership_dynamic,),
    # Output, relay, binsensors, ... status commands
    "SMA1": (PckGenerator.request_output_status, 0),
    "SMA2": (PckGenerator.request_output_status, 1),
    "SMA3": (PckGenerator.request_output_status, 2),
    "SMA4": (PckGenerator.request_output_status, 3),
    "SMR": (PckGenerator.request_relays_status,),
    "SMB": (PckGenerator.request_bin_sensors_status,),
    "SMT": (PckGenerator.request_leds_and_logic_ops,),
    "STX": (PckGenerator.request_key_lock_status,),
    # Variable status (new commands)
    **{
        f"MWT{Var.to_var_id(var) + 1}": (
            PckGenerator.request_var_status,
            var,
            NEW_VAR_SW_AGE,
        )
        for var in Var.variables()
    },
    **{
        f"MWS{Var.to_set_point_id(var) + 1}": (
            PckGenerator.request_var_status,
            var,
            NEW_VAR_SW_AGE,
        )
        for var in Var.set_points()
    },
    **{
        f"MWC{Var.to_s0_id(var) + 1}": (
            PckGenerator.request_var_status,
            var,
            NEW_VAR_SW_AGE,
        )
        for var in Var.s0s()
    },
    **{
        f"SE{Var.to_thrs_register_id(var) + 1}": (
            PckGenerator.request_var_status,
            var,
            NEW_VAR_SW_AGE,
        )
        for reg in Var.thresholds()
        for var in reg
    },
    # Variable status (legacy commands)
    "MWV": (PckGenerator.request_var_status, Var.TVAR, NEW_VAR_SW_AGE - 1),
    "MWTA": (PckGenerator.request_var_status, Var.R1VAR, NEW_VAR_SW_AGE - 1),
    "MWTB": (PckGenerator.request_var_status, Var.R2VAR, NEW_VAR_SW_AGE - 1),
    "MWSA": (PckGenerator.request_var_status, Var.R1VARSETPOINT, NEW_VAR_SW_AGE - 1),
    "MWSB": (PckGenerator.request_var_status, Var.R2VARSETPOINT, NEW_VAR_SW_AGE - 1),
    **{
        "SL1": (PckGenerator.request_var_status, var, NEW_VAR_SW_AGE - 1)
        for var in Var.thresholds()[0]
    },
    # Output manipulation
    **{
        f"A{output + 1:d}DI050123": (PckGenerator.dim_output, output, 50.0, 123)
        for output in range(4)
    },
    **{
        f"O{output + 1:d}DI101123": (PckGenerator.dim_output, output, 50.5, 123)
        for output in range(4)
    },
    "OY100100100100123": (PckGenerator.dim_all_outputs, 50.0, 123, 0x180501),
    "OY000000000000123": (PckGenerator.dim_all_outputs, 0.0, 123, 0x180501),
    "OY200200200200123": (PckGenerator.dim_all_outputs, 100.0, 123, 0x180501),
    "AA123": (PckGenerator.dim_all_outputs, 0.0, 123, 0x180500),
    "AE123": (PckGenerator.dim_all_outputs, 100.0, 123, 0x180500),
    "AH050": (PckGenerator.dim_all_outputs, 50.0, 123, 0x180500),
    **{
        f"A{output + 1:d}AD050": (PckGenerator.rel_output, output, 50.0)
        for output in range(4)
    },
    **{
        f"A{output + 1:d}SB050": (PckGenerator.rel_output, output, -50.0)
        for output in range(4)
    },
    **{
        f"O{output + 1:d}AD101": (PckGenerator.rel_output, output, 50.5)
        for output in range(4)
    },
    **{
        f"O{output + 1:d}SB101": (PckGenerator.rel_output, output, -50.5)
        for output in range(4)
    },
    **{
        f"A{output + 1:d}TA123": (PckGenerator.toggle_output, output, 123)
        for output in range(4)
    },
    **{
        f"A{output + 1:d}MT123": (PckGenerator.toggle_output, output, 123, True)
        for output in range(4)
    },
    "AU123": (PckGenerator.toggle_all_outputs, 123),
    # Relay state manipulation
    "R80-1U1-U0": (
        PckGenerator.control_relays,
        [
            RelayStateModifier.OFF,
            RelayStateModifier.NOCHANGE,
            RelayStateModifier.ON,
            RelayStateModifier.TOGGLE,
            RelayStateModifier.ON,
            RelayStateModifier.NOCHANGE,
            RelayStateModifier.TOGGLE,
            RelayStateModifier.OFF,
        ],
    ),
    "R8T0321--111--": (
        PckGenerator.control_relays_timer,
        30 * 32,
        [
            RelayStateModifier.ON,
            RelayStateModifier.NOCHANGE,
            RelayStateModifier.NOCHANGE,
            RelayStateModifier.ON,
            RelayStateModifier.ON,
            RelayStateModifier.ON,
            RelayStateModifier.NOCHANGE,
            RelayStateModifier.NOCHANGE,
        ],
    ),
    # Motor state manipulation
    "R8--10----": (
        PckGenerator.control_motor_relays,
        1,
        MotorStateModifier.UP,
    ),
    "R8-----U--": (
        PckGenerator.control_motor_relays,
        2,
        MotorStateModifier.TOGGLEDIR,
    ),
    "R8UU------": (
        PckGenerator.control_motor_relays,
        0,
        MotorStateModifier.CYCLE,
    ),
    "R8M1ZU": (
        PckGenerator.control_motor_relays,
        0,
        MotorStateModifier.UP,
        MotorPositioningMode.BS4,
    ),
    "R8M2AU": (
        PckGenerator.control_motor_relays,
        1,
        MotorStateModifier.DOWN,
        MotorPositioningMode.BS4,
    ),
    "R8M5ST": (
        PckGenerator.control_motor_relays,
        2,
        MotorStateModifier.STOP,
        MotorPositioningMode.BS4,
    ),
    "R8M1GP200": (
        PckGenerator.control_motor_relays_position,
        0,
        0.0,
        MotorPositioningMode.BS4,
    ),
    "R8M6GP100": (
        PckGenerator.control_motor_relays_position,
        3,
        50.0,
        MotorPositioningMode.BS4,
    ),
    "R8M3P1": (
        PckGenerator.request_motor_position_status,
        0,
    ),
    "R8M7P2": (
        PckGenerator.request_motor_position_status,
        1,
    ),
    "JH050001": (
        PckGenerator.control_motor_relays_position,
        0,
        50,
        MotorPositioningMode.MODULE,
    ),
    "JH030004": (
        PckGenerator.control_motor_relays_position,
        2,
        70,
        MotorPositioningMode.MODULE,
    ),
    "X2001228000": (
        PckGenerator.control_motor_outputs,
        MotorStateModifier.UP,
        MotorReverseTime.RT70,
    ),
    "A1DI100008": (
        PckGenerator.control_motor_outputs,
        MotorStateModifier.UP,
        MotorReverseTime.RT600,
    ),
    "A1DI100011": (
        PckGenerator.control_motor_outputs,
        MotorStateModifier.UP,
        MotorReverseTime.RT1200,
    ),
    "X2001000228": (
        PckGenerator.control_motor_outputs,
        MotorStateModifier.DOWN,
        MotorReverseTime.RT70,
    ),
    "A2DI100008": (
        PckGenerator.control_motor_outputs,
        MotorStateModifier.DOWN,
        MotorReverseTime.RT600,
    ),
    "A2DI100011": (
        PckGenerator.control_motor_outputs,
        MotorStateModifier.DOWN,
        MotorReverseTime.RT1200,
    ),
    "AY000000": (
        PckGenerator.control_motor_outputs,
        MotorStateModifier.STOP,
    ),
    "JE": (
        PckGenerator.control_motor_outputs,
        MotorStateModifier.CYCLE,
    ),
    # Variable manipulation
    **{
        f"X2{var.value | 0x40:03d}016225": (PckGenerator.update_status_var, var, 4321)
        for var in Var.variables()
    },
    "X2030044129": (PckGenerator.var_abs, Var.R1VARSETPOINT, 4201),
    "X2030108129": (PckGenerator.var_abs, Var.R2VARSETPOINT, 4201),
    "X2030032000": (PckGenerator.var_reset, Var.R1VARSETPOINT, 0x170206),
    "X2030096000": (PckGenerator.var_reset, Var.R2VARSETPOINT, 0x170206),
    "ZS30000": (PckGenerator.var_reset, Var.TVAR, 0x170205),
    **{
        f"Z-{var.value + 1:03d}4090": (PckGenerator.var_reset, var, 0x170206)
        for var in Var.variables()
    },
    "ZA23423": (PckGenerator.var_rel, Var.TVAR, RelVarRef.CURRENT, 23423, 0x170205),
    "ZS23423": (PckGenerator.var_rel, Var.TVAR, RelVarRef.CURRENT, -23423, 0x170205),
    **{
        f"Z-{var.value + 1:03d}3000": (
            PckGenerator.var_rel,
            var,
            RelVarRef.CURRENT,
            -3000,
            0x170206,
        )
        for var in Var.variables()
        if var != Var.TVAR
    },
    **{
        f"RE{('A', 'B')[nvar]}S{('A', 'P')[nref]}-500": (
            PckGenerator.var_rel,
            var,
            ref,
            -500,
            sw_age,
        )
        for nvar, var in enumerate(Var.set_points())
        for nref, ref in enumerate(RelVarRef)
        for sw_age in (0x170206, 0x170205)
    },
    **{
        f"RE{('A', 'B')[nvar]}S{('A', 'P')[nref]}+500": (
            PckGenerator.var_rel,
            var,
            ref,
            500,
            sw_age,
        )
        for nvar, var in enumerate(Var.set_points())
        for nref, ref in enumerate(RelVarRef)
        for sw_age in (0x170206, 0x170205)
    },
    **{
        f"SS{('R', 'E')[nref]}0500SR{r + 1}{i + 1}": (
            PckGenerator.var_rel,
            Var.thresholds()[r][i],
            ref,
            -500,
            0x170206,
        )
        for r in range(4)
        for i in range(4)
        for nref, ref in enumerate(RelVarRef)
    },
    **{
        f"SS{('R', 'E')[nref]}0500AR{r + 1}{i + 1}": (
            PckGenerator.var_rel,
            Var.thresholds()[r][i],
            ref,
            500,
            0x170206,
        )
        for r in range(4)
        for i in range(4)
        for nref, ref in enumerate(RelVarRef)
    },
    **{
        f"SS{('R', 'E')[nref]}0500S{1 << (4 - i):05b}": (
            PckGenerator.var_rel,
            Var.thresholds()[0][i],
            ref,
            -500,
            0x170205,
        )
        for i in range(5)
        for nref, ref in enumerate(RelVarRef)
    },
    **{
        f"SS{('R', 'E')[nref]}0500A{1 << (4 - i):05b}": (
            PckGenerator.var_rel,
            Var.thresholds()[0][i],
            ref,
            500,
            0x170205,
        )
        for i in range(5)
        for nref, ref in enumerate(RelVarRef)
    },
    # Led manipulation
    **{
        f"LA{led + 1:03d}{state.value}": (PckGenerator.control_led, led, state)
        for led in range(12)
        for state in LedStatus
    },
    # Send keys
    **{
        f"TS{acmd.value}{bcmd.value}{ccmd.value}10011100": (
            PckGenerator.send_keys,
            [acmd, bcmd, ccmd, SendKeyCommand.DONTSEND],
            [True, False, False, True, True, True, False, False],
        )
        for acmd in SendKeyCommand
        for bcmd in SendKeyCommand
        for ccmd in SendKeyCommand
    },
    **{
        f"TS---{dcmd.value}10011100": (
            PckGenerator.send_keys,
            [
                SendKeyCommand.DONTSEND,
                SendKeyCommand.DONTSEND,
                SendKeyCommand.DONTSEND,
                dcmd,
            ],
            [True, False, False, True, True, True, False, False],
        )
        for dcmd in SendKeyCommand
        if dcmd != SendKeyCommand.DONTSEND
    },
    **{
        f"TV{('A', 'B', 'C', 'D')[table]}040{unit.value}11001110": (
            PckGenerator.send_keys_hit_deferred,
            table,
            40,
            unit,
            [True, True, False, False, True, True, True, False],
        )
        for table in range(4)
        for unit in TimeUnit
    },
    # Lock keys
    **{
        f"TX{('A', 'B', 'C', 'D')[table]}10U--01U": (
            PckGenerator.lock_keys,
            table,
            [
                KeyLockStateModifier.ON,
                KeyLockStateModifier.OFF,
                KeyLockStateModifier.TOGGLE,
                KeyLockStateModifier.NOCHANGE,
                KeyLockStateModifier.NOCHANGE,
                KeyLockStateModifier.OFF,
                KeyLockStateModifier.ON,
                KeyLockStateModifier.TOGGLE,
            ],
        )
        for table in range(4)
    },
    **{
        f"TXZA040{unit.value}11001110": (
            PckGenerator.lock_keys_tab_a_temporary,
            40,
            unit,
            [
                KeyLockStateModifier.ON,
                KeyLockStateModifier.ON,
                KeyLockStateModifier.OFF,
                KeyLockStateModifier.OFF,
                KeyLockStateModifier.ON,
                KeyLockStateModifier.ON,
                KeyLockStateModifier.ON,
                KeyLockStateModifier.OFF,
            ],
        )
        for unit in TimeUnit
    },
    # Lock regulator
    **{
        f"RE{('A', 'B')[reg]:s}XS": (PckGenerator.lock_regulator, reg, True, -1)
        for reg in range(2)
    },
    **{
        f"RE{('A', 'B')[reg]:s}XA": (PckGenerator.lock_regulator, reg, False, -1)
        for reg in range(2)
    },
    **{
        f"X2030{0x40 * reg + 0x07:03d}{2 * value:03d}": (
            PckGenerator.lock_regulator,
            reg,
            True,
            0x120301,
            value,
        )
        for reg in range(2)
        for value in (0, 50, 100)
    },
    **{
        f"RE{('A', 'B')[reg]:s}XS": (PckGenerator.lock_regulator, reg, True, 0x120301)
        for reg in range(2)
    },
    **{
        f"RE{('A', 'B')[reg]:s}XA": (PckGenerator.lock_regulator, reg, False, 0x120301)
        for reg in range(2)
    },
    # Lock thresholds
    **{
        f"SX{reg + 1:d}10-1": (
            PckGenerator.lock_thresholds,
            reg,
            [
                ThresholdLockStateModifier.ON,
                ThresholdLockStateModifier.OFF,
                ThresholdLockStateModifier.NOCHANGE,
                ThresholdLockStateModifier.ON,
            ],
        )
        for reg in range(4)
    },
    # scenes
    "SZR003007": (PckGenerator.request_status_scene, 3, 7),
    "SZD003007200100101050": (
        PckGenerator.store_scene_outputs_direct,
        3,
        7,
        (100, 50.5),
        (100, 50),
    ),
    "SZD003007200100101050021000199107": (
        PckGenerator.store_scene_outputs_direct,
        3,
        7,
        (100, 50.5, 10.5, 99.5),
        (100, 50, 0, 107),
    ),
    "SZW004": (PckGenerator.change_scene_register, 4),
    "SZA7001": (PckGenerator.activate_scene_output, 1, OutputPort),
    "SZA7001133": (PckGenerator.activate_scene_output, 1, OutputPort, 133),
    "SZS7002": (PckGenerator.store_scene_output, 2, OutputPort),
    "SZS7002133": (PckGenerator.store_scene_output, 2, OutputPort, 133),
    "SZA7005": (PckGenerator.activate_scene_output, 5, OutputPort),
    "SZA7005133": (PckGenerator.activate_scene_output, 5, OutputPort, 133),
    "SZS7008": (PckGenerator.store_scene_output, 8, OutputPort),
    "SZS7008133": (PckGenerator.store_scene_output, 8, OutputPort, 133),
    "SZA000810001110": (
        PckGenerator.activate_scene_relay,
        8,
        (
            RelayPort.RELAY1,
            RelayPort.RELAY5,
            RelayPort.RELAY6,
            RelayPort.RELAY7,
        ),
    ),
    "SZS000810001110": (
        PckGenerator.store_scene_relay,
        8,
        (
            RelayPort.RELAY1,
            RelayPort.RELAY5,
            RelayPort.RELAY6,
            RelayPort.RELAY7,
        ),
    ),
    # dynamic text
    **{
        f"GTDT{row + 1:d}{part + 1:d}asdfasdfasdf".encode(): (
            PckGenerator.dyn_text_part,
            row,
            part,
            b"asdfasdfasdf",
        )
        for row in range(4)
        for part in range(5)
    },
    b"GTDT45\xff\xfe\x80\x34\xdd\xcc\xaa\xbf\x00\xac": (
        PckGenerator.dyn_text_part,
        3,
        4,
        b"\xff\xfe\x80\x34\xdd\xcc\xaa\xbf\x00\xac",
    ),
}


@pytest.mark.parametrize("expected, command", COMMANDS.items())
def test_command_generation_single_mod_noack(
    expected: str, command: tuple[Any, ...]
) -> None:
    """Test if InputMod parses message correctly."""
    assert expected == command[0](*command[1:])



"""Init file for pypck."""

from pypck import (
    connection,
    device,
    helpers,
    inputs,
    lcn_addr,
    lcn_defs,
    pck_commands,
    status_requester,
)

__all__ = [
    "connection",
    "device",
    "helpers",
    "inputs",
    "lcn_addr",
    "lcn_defs",
    "pck_commands",
    "status_requester",
]



"""Parsers for incoming PCK messages."""

from __future__ import annotations

import logging

from pypck import lcn_defs
from pypck.lcn_addr import LcnAddr
from pypck.pck_commands import PckParser

_LOGGER = logging.getLogger(__name__)


class Input:
    """Parent class for all input data read from LCN-PCHK.

    An implementation of :class:`~pypck.input.Input` has to provide easy
    accessible attributes and/or methods to expose the PCK command properties
    to the user.
    Each Input object provides an implementation of
    :func:`~pypck.input.Input.try_parse` static method, to parse the given
    plain text PCK command. If the command can be parsed by the Input object,
    a list of instances of :class:`~pypck.input.Input` is returned. Otherwise,
    nothing is returned.
    """

    def __init__(self) -> None:
        """Construct Input object."""
        self.pck = ""

    @staticmethod
    def try_parse(data: str) -> list[Input] | None:
        """Try to parse the given input text.

        Will return a list of parsed Inputs. The list might be empty (but not
        null).
        """
        raise NotImplementedError


class ModInput(Input):  # pylint: disable=too-few-public-methods, abstract-method
    """Parent class of all inputs having an LCN module as its source.

    The class in inherited from :class:`~pypck.input.Input`
    """

    def __init__(self, physical_source_addr: LcnAddr):
        """Construct ModInput object."""
        super().__init__()
        self.physical_source_addr = physical_source_addr


# ## Plain text inputs


class AuthUsername(Input):
    """Authentication username message received from PCHK."""

    @staticmethod
    def try_parse(data: str) -> list[Input] | None:
        """Try to parse the given input text.

        Will return a list of parsed Inputs. The list might be empty (but not
        null).

        :param    data    str:    The input data received from LCN-PCHK

        :return:            The parsed Inputs
        :rtype:             List with instances of :class:`~pypck.input.Input`
        """
        if data == PckParser.AUTH_USERNAME:
            return [AuthUsername()]
        return None


class AuthPassword(Input):
    """Authentication password message received from PCHK."""

    @staticmethod
    def try_parse(data: str) -> list[Input] | None:
        """Try to parse the given input text.

        Will return a list of parsed Inputs. The list might be empty (but not
        null).

        :param    data    str:    The input data received from LCN-PCHK

        :return:            The parsed Inputs
        :rtype:             List with instances of :class:`~pypck.input.Input`
        """
        if data == PckParser.AUTH_PASSWORD:
            return [AuthPassword()]
        return None


class AuthOk(Input):
    """Authentication ok message received from PCHK."""

    @staticmethod
    def try_parse(data: str) -> list[Input] | None:
        """Try to parse the given input text.

        Will return a list of parsed Inputs. The list might be empty (but not
        null).

        :param    data    str:    The input data received from LCN-PCHK

        :return:            The parsed Inputs (never null)
        :rtype:             List with instances of :class:`~pypck.input.Input`
        """
        if data == PckParser.AUTH_OK:
            return [AuthOk()]
        return None


class AuthFailed(Input):
    """Authentication failed message received from PCHK."""

    @staticmethod
    def try_parse(data: str) -> list[Input] | None:
        """Try to parse the given input text.

        Will return a list of parsed Inputs. The list might be empty (but not
        null).

        :param    data    str:    The input data received from LCN-PCHK

        :return:            The parsed Inputs (never null)
        :rtype:             List with instances of :class:`~pypck.input.Input`
        """
        if data == PckParser.AUTH_FAILED:
            return [AuthFailed()]
        return None


class LcnConnState(Input):
    """LCN bus connected message received from PCHK."""

    def __init__(self, is_lcn_connected: bool):
        """Construct Input object."""
        super().__init__()
        self._is_lcn_connected = is_lcn_connected

    @property
    def is_lcn_connected(self) -> bool:
        """Return the LCN bus connection status.

        :return:   True if connection to hardware bus was established,
                   otherwise False.
        :rtype:    bool
        """
        return self._is_lcn_connected

    @staticmethod
    def try_parse(data: str) -> list[Input] | None:
        """Try to parse the given input text.

        Will return a list of parsed Inputs. The list might be empty (but not
        null).

        :param    data    str:    The input data received from LCN-PCHK

        :return:            The parsed Inputs (never null)
        :rtype:             List with instances of :class:`~pypck.input.Input`
        """
        if data == PckParser.LCNCONNSTATE_CONNECTED:
            return [LcnConnState(True)]
        if data == PckParser.LCNCONNSTATE_DISCONNECTED:
            return [LcnConnState(False)]
        return None


class LicenseError(Input):
    """LCN bus connected message received from PCHK."""

    @staticmethod
    def try_parse(data: str) -> list[Input] | None:
        """Try to parse the given input text.

        Will return a list of parsed Inputs. The list might be empty (but not
        null).

        :param    data    str:    The input data received from LCN-PCHK

        :return:            The parsed Inputs (never null)
        :rtype:             List with instances of :class:`~pypck.input.Input`
        """
        if data == PckParser.LICENSE_ERROR:
            return [LicenseError()]
        return None


class DecModeSet(Input):
    """Decimal mode set received from PCHK."""

    @staticmethod
    def try_parse(data: str) -> list[Input] | None:
        """Try to parse the given input text.

        Will return a list of parsed Inputs. The list might be empty (but not
        null).

        :param    data    str:    The input data received from LCN-PCHK

        :return:            The parsed Inputs (never null)
        :rtype:             List with instances of :class:`~pypck.input.Input`
        """
        if data == PckParser.DEC_MODE_SET:
            return [DecModeSet()]
        return None


class CommandError(Input):
    """Command error received from PCHK."""

    def __init__(self, message: str):
        """Construct Input object."""
        super().__init__()
        self.message = message

    @staticmethod
    def try_parse(data: str) -> list[Input] | None:
        """Try to parse the given input text.

        Will return a list of parsed Inputs. The list might be empty (but not
        null).

        :param    data    str:    The input data received from LCN-PCHK

        :return:            The parsed Inputs (never null)
        :rtype:             List with instances of :class:`~pypck.input.Input`
        """
        matcher = PckParser.PATTERN_COMMAND_ERROR.match(data)
        if matcher:
            return [CommandError(matcher.group("message"))]
        return None


class Ping(Input):
    """Ping message received from PCHK."""

    def __init__(self, count: int | None):
        """Construct Input object."""
        super().__init__()
        self._count = count

    @property
    def count(self) -> int | None:
        """Return the ping count.

        :return:   Ping count
        :rtype:    int | None
        """
        return self._count

    @staticmethod
    def try_parse(data: str) -> list[Input] | None:
        """Try to parse the given input text.

        Will return a list of parsed Inputs. The list might be empty (but not
        null).

        :param    data    str:    The input data received from LCN-PCHK

        :return:            The parsed Inputs (never null)
        :rtype:             List with instances of :class:`~pypck.input.Input`
        """
        matcher = PckParser.PATTERN_PING.match(data)
        if matcher:
            count = matcher.group("count")
            if count == "":
                return [Ping(None)]
            return [Ping(int(count))]

        return None


# ## Inputs received from modules


class ModAck(ModInput):
    """Acknowledge message received from module."""

    def __init__(
        self, physical_source_addr: LcnAddr, code: lcn_defs.AcknowledgeErrorCode
    ):
        """Construct ModInput object."""
        super().__init__(physical_source_addr)
        self.code = code

    def get_code(self) -> lcn_defs.AcknowledgeErrorCode:
        """Return the acknowledge code."""
        return self.code

    @staticmethod
    def try_parse(data: str) -> list[Input] | None:
        """Try to parse the given input text.

        Will return a list of parsed Inputs. The list might be empty (but not
        null).

        :param    data    str:    The input data received from LCN-PCHK

        :return:            The parsed Inputs (never null)
        :rtype:             List with instances of :class:`~pypck.input.Input`
        """
        matcher_pos = PckParser.PATTERN_ACK_POS.match(data)
        if matcher_pos:
            addr = LcnAddr(
                int(matcher_pos.group("seg_id")), int(matcher_pos.group("mod_id"))
            )
            return [ModAck(addr, lcn_defs.AcknowledgeErrorCode.OK)]

        matcher_neg = PckParser.PATTERN_ACK_NEG.match(data)
        if matcher_neg:
            addr = LcnAddr(
                int(matcher_neg.group("seg_id")), int(matcher_neg.group("mod_id"))
            )
            return [
                ModAck(
                    addr, lcn_defs.AcknowledgeErrorCode(int(matcher_neg.group("code")))
                )
            ]

        return None


class ModSk(ModInput):
    """Segment information received from an LCN segment coupler."""

    def __init__(self, physical_source_addr: LcnAddr, reported_seg_id: int):
        """Construct ModInput object."""
        super().__init__(physical_source_addr)
        self.reported_seg_id = reported_seg_id

    def get_reported_seg_id(self) -> int:
        """Return the segment id reported from segment coupler.

        :return:   Reported segment id.
        :rtype:    int
        """
        return self.reported_seg_id

    @staticmethod
    def try_parse(data: str) -> list[Input] | None:
        """Try to parse the given input text.

        Will return a list of parsed Inputs. The list might be empty (but not
        null).

        :param    data    str:    The input data received from LCN-PCHK

        :return:            The parsed Inputs (never null)
        :rtype:             List with instances of :class:`~pypck.input.Input`
        """
        matcher = PckParser.PATTERN_SK_RESPONSE.match(data)
        if matcher:
            addr = LcnAddr(int(matcher.group("seg_id")), int(matcher.group("mod_id")))
            return [ModSk(addr, int(matcher.group("id")))]

        return None


class ModSn(ModInput):
    """Serial number and firmware version received from an LCN module."""

    def __init__(
        self,
        physical_source_addr: LcnAddr,
        hardware_serial: int,
        manu: int,
        software_serial: int,
        hardware_type: lcn_defs.HardwareType,
    ):
        """Construct ModInput object."""
        super().__init__(physical_source_addr)
        self.hardware_serial = hardware_serial
        self.manu = manu
        self.software_serial = software_serial
        self.hardware_type = hardware_type

    @staticmethod
    def try_parse(data: str) -> list[Input] | None:
        """Try to parse the given input text.

        Will return a list of parsed Inputs. The list might be empty (but not
        null).

        :param    data    str:    The input data received from LCN-PCHK

        :return:            The parsed Inputs (never null)
        :rtype:             List with instances of :class:`~pypck.input.Input`
        """
        matcher = PckParser.PATTERN_SN.match(data)
        if matcher:
            addr = LcnAddr(int(matcher.group("seg_id")), int(matcher.group("mod_id")))
            hardware_serial = int(matcher.group("hardware_serial"), 16)
            try:
                manu = int(matcher.group("manu"), 16)
            except (
                ValueError
            ):  # unconventional manufacturer code (e.g., due to LinHK VM)
                manu = 0xFF
                _LOGGER.warning(
                    "Unconventional manufacturer code for module (S%d, M%d): %s. Defaulting to 0x%02X",
                    addr.seg_id,
                    addr.addr_id,
                    matcher.group("manu"),
                    manu,
                )
            software_serial = int(matcher.group("software_serial"), 16)
            try:
                hardware_type = lcn_defs.HardwareType(
                    int(matcher.group("hardware_type"))
                )
            except ValueError:  # unknown hardware type
                hardware_type = lcn_defs.HardwareType(-1)
            return [ModSn(addr, hardware_serial, manu, software_serial, hardware_type)]

        return None


class ModNameComment(ModInput):
    """Name or comment received from an LCN module."""

    def __init__(
        self, physical_source_addr: LcnAddr, command: str, block_id: int, text: str
    ):
        """Construct ModInput object."""
        super().__init__(physical_source_addr)
        self.command = command
        self.block_id = block_id
        self.text = text

    @staticmethod
    def try_parse(data: str) -> list[Input] | None:
        """Try to parse the given input text.

        Will return a list of parsed Inputs. The list might be empty (but not
        null).

        :param    data    str:    The input data received from LCN-PCHK

        :return:            The parsed Inputs (never null)
        :rtype:             List with instances of :class:`~pypck.input.Input`
        """
        matcher = PckParser.PATTERN_NAME_COMMENT.match(data)
        if matcher:
            addr = LcnAddr(int(matcher.group("seg_id")), int(matcher.group("mod_id")))
            command = matcher.group("command")
            block_id = int(matcher.group("block_id")) - 1
            text = matcher.group("text")
            return [ModNameComment(addr, command, block_id, text)]

        return None


class ModStatusGroups(ModInput):
    """Group memberships status received from an LCN module."""

    def __init__(
        self,
        physical_source_addr: LcnAddr,
        dynamic: bool,
        max_groups: int,
        groups: list[LcnAddr],
    ):
        """Construct ModInput object."""
        super().__init__(physical_source_addr)
        self.dynamic = dynamic
        self.max_groups = max_groups
        self.groups = groups

    @staticmethod
    def try_parse(data: str) -> list[Input] | None:
        """Try to parse the given input text.

        Will return a list of parsed Inputs. The list might be empty (but not
        null).

        :param    data    str:    The input data received from LCN-PCHK

        :return:            The parsed Inputs (never null)
        :rtype:             List with instances of :class:`~pypck.input.Input`
        """
        matcher = PckParser.PATTERN_STATUS_GROUPS.match(data)
        if matcher:
            addr = LcnAddr(int(matcher.group("seg_id")), int(matcher.group("mod_id")))
            dynamic = matcher.group("kind") == "D"
            max_groups = int(matcher.group("max_groups"))
            groups = [
                LcnAddr(addr.seg_id, int(group), True)
                for group in matcher.groups()[4:]
                if group is not None
            ]
            return [ModStatusGroups(addr, dynamic, max_groups, groups)]

        return None


class ModStatusOutput(ModInput):
    """Status of an output-port in percent received from an LCN module."""

    def __init__(self, physical_source_addr: LcnAddr, output_id: int, percent: float):
        """Construct ModInput object."""
        super().__init__(physical_source_addr)
        self.output_id = output_id
        self.percent = percent

    def get_output_id(self) -> int:
        """Return the output port id.

        :return:    Output port id.
        :rtype:     int
        """
        return self.output_id

    def get_percent(self) -> float:
        """Return the output brightness in percent.

        :return:    Brightness in percent.
        :rtype:     float
        """
        return self.percent

    @staticmethod
    def try_parse(data: str) -> list[Input] | None:
        """Try to parse the given input text.

        Will return a list of parsed Inputs. The list might be empty (but not
        null).

        :param    data    str:    The input data received from LCN-PCHK

        :return:            The parsed Inputs (never null)
        :rtype:             List with instances of :class:`~pypck.input.Input`
        """
        matcher = PckParser.PATTERN_STATUS_OUTPUT_PERCENT.match(data)
        if matcher:
            addr = LcnAddr(int(matcher.group("seg_id")), int(matcher.group("mod_id")))
            return [
                ModStatusOutput(
                    addr,
                    int(matcher.group("output_id")) - 1,
                    float(matcher.group("percent")),
                )
            ]

        return None


class ModStatusOutputNative(ModInput):
    """Status of an output-port in native units received from an LCN module."""

    def __init__(self, physical_source_addr: LcnAddr, output_id: int, value: int):
        """Construct ModInput object."""
        super().__init__(physical_source_addr)
        self.output_id = output_id
        self.value = value

    def get_output_id(self) -> int:
        """Return the output port id.

        :return:    Output port id.
        :rtype:     int
        """
        return self.output_id

    def get_value(self) -> int:
        """Return the output brightness in native units.

        :return:    Brightness in percent.
        :rtype:     float
        """
        return self.value

    @staticmethod
    def try_parse(data: str) -> list[Input] | None:
        """Try to parse the given input text.

        Will return a list of parsed Inputs. The list might be empty (but not
        null).

        :param    data    str:    The input data received from LCN-PCHK

        :return:            The parsed Inputs (never null)
        :rtype:             List with instances of :class:`~pypck.input.Input`
        """
        matcher = PckParser.PATTERN_STATUS_OUTPUT_NATIVE.match(data)
        if matcher:
            addr = LcnAddr(int(matcher.group("seg_id")), int(matcher.group("mod_id")))
            return [
                ModStatusOutputNative(
                    addr,
                    int(matcher.group("output_id")) - 1,
                    int(matcher.group("value")),
                )
            ]

        return None


class ModStatusRelays(ModInput):
    """Status of 8 relays received from an LCN module.

    Includes helper functions for motor states based on LCN wiring.
    """

    def __init__(self, physical_source_addr: LcnAddr, states: list[bool]):
        """Construct ModInput object."""
        super().__init__(physical_source_addr)
        self.states = states

    def get_state(self, relay_id: int) -> bool:
        """Get the state of a single relay.

        :param    int    relay_id:    Relay id (0..7)

        :return:                      The relay's state
        :rtype:   bool
        """
        return self.states[relay_id]

    def get_motor_onoff_relay(self, motor_id: int) -> int:
        """Get the motor on/off relay id."""
        if 0 > motor_id > 3:
            raise ValueError("Motor id must be in range 0..3")
        return motor_id * 2

    def get_motor_updown_relay(self, motor_id: int) -> int:
        """Get the motor up/down relay id."""
        if 0 > motor_id > 3:
            raise ValueError("Motor id must be in range 0..3")
        return motor_id * 2 + 1

    def motor_is_on(self, motor_id: int) -> bool:
        """Check if a motor is on."""
        return self.states[self.get_motor_onoff_relay(motor_id)]

    def is_opening(self, motor_id: int) -> bool:
        """Check if a motor is opening."""
        if self.motor_is_on(motor_id):
            return not self.states[self.get_motor_updown_relay(motor_id)]
        return False

    def is_closing(self, motor_id: int) -> bool:
        """Check if a motor is closing."""
        if self.motor_is_on(motor_id):
            return self.states[self.get_motor_updown_relay(motor_id)]
        return False

    def is_assumed_closed(self, motor_id: int) -> bool:
        """Check if a motor is closed.

        The closed state is assumed if the motor direction is down and the motor is switched off."
        """
        if not self.motor_is_on(motor_id):
            return self.states[self.get_motor_updown_relay(motor_id)]
        return False

    @staticmethod
    def try_parse(data: str) -> list[Input] | None:
        """Try to parse the given input text.

        Will return a list of parsed Inputs. The list might be empty (but not
        null).

        :param    data    str:    The input data received from LCN-PCHK

        :return:            The parsed Inputs (never null)
        :rtype:             List with instances of :class:`~pypck.input.Input`
        """
        matcher = PckParser.PATTERN_STATUS_RELAYS.match(data)
        if matcher:
            addr = LcnAddr(int(matcher.group("seg_id")), int(matcher.group("mod_id")))
            return [
                ModStatusRelays(
                    addr, PckParser.get_boolean_value(int(matcher.group("byte_value")))
                )
            ]

        return None


class ModStatusBinSensors(ModInput):
    """Status of 8 binary sensors received from an LCN module."""

    def __init__(self, physical_source_addr: LcnAddr, states: list[bool]):
        """Construct ModInput object."""
        super().__init__(physical_source_addr)
        self.states = states

    def get_state(self, bin_sensor_id: int) -> bool:
        """Get the state of a single binary-sensor.

        :param    int    bin_sensor_id:    Binary sensor id (0..7)

        :return:                           The binary-sensor's state
        :rtype:   bool
        """
        return self.states[bin_sensor_id]

    @staticmethod
    def try_parse(data: str) -> list[Input] | None:
        """Try to parse the given input text.

        Will return a list of parsed Inputs. The list might be empty (but not
        null).

        :param    data    str:    The input data received from LCN-PCHK

        :return:            The parsed Inputs (never null)
        :rtype:             List with instances of :class:`~pypck.input.Input`
        """
        matcher = PckParser.PATTERN_STATUS_BINSENSORS.match(data)
        if matcher:
            addr = LcnAddr(int(matcher.group("seg_id")), int(matcher.group("mod_id")))
            return [
                ModStatusBinSensors(
                    addr, PckParser.get_boolean_value(int(matcher.group("byte_value")))
                )
            ]

        return None


class ModStatusVar(ModInput):
    """Status of a variable received from an LCN module."""

    def __init__(
        self,
        physical_source_addr: LcnAddr,
        orig_var: lcn_defs.Var,
        value: lcn_defs.VarValue,
    ):
        """Construct ModInput object."""
        super().__init__(physical_source_addr)
        self.orig_var = orig_var
        self.value = value
        self.var = self.orig_var

    def get_var(self) -> lcn_defs.Var:
        """Get the variable's real type.

        :return:        The real type
        :rtype:        :class:`~pypck.lcn_defs.Var`
        """
        return self.var

    def get_value(self) -> lcn_defs.VarValue:
        """Get the variable's value.

        :return:    The value of the variable.
        :rtype:     int
        """
        return self.value

    @staticmethod
    def try_parse(data: str) -> list[Input] | None:
        """Try to parse the given input text.

        Will return a list of parsed Inputs. The list might be empty (but not
        null).

        :param    data    str:    The input data received from LCN-PCHK

        :return:            The parsed Inputs (never null)
        :rtype:             List with instances of :class:`~pypck.input.Input`
        """
        matcher = PckParser.PATTERN_STATUS_VAR.match(data)
        if matcher:
            addr = LcnAddr(int(matcher.group("seg_id")), int(matcher.group("mod_id")))
            var = lcn_defs.Var.var_id_to_var(int(matcher.group("id")) - 1)
            value = lcn_defs.VarValue.from_native(int(matcher.group("value")))
            return [ModStatusVar(addr, var, value)]

        matcher = PckParser.PATTERN_STATUS_SETVAR.match(data)
        if matcher:
            addr = LcnAddr(int(matcher.group("seg_id")), int(matcher.group("mod_id")))
            var = lcn_defs.Var.set_point_id_to_var(int(matcher.group("id")) - 1)
            value = lcn_defs.VarValue.from_native(int(matcher.group("value")))
            return [ModStatusVar(addr, var, value)]

        matcher = PckParser.PATTERN_STATUS_THRS.match(data)
        if matcher:
            addr = LcnAddr(int(matcher.group("seg_id")), int(matcher.group("mod_id")))
            var = lcn_defs.Var.thrs_id_to_var(
                int(matcher.group("register_id")) - 1, int(matcher.group("thrs_id")) - 1
            )
            value = lcn_defs.VarValue.from_native(int(matcher.group("value")))
            return [ModStatusVar(addr, var, value)]

        matcher = PckParser.PATTERN_STATUS_S0INPUT.match(data)
        if matcher:
            addr = LcnAddr(int(matcher.group("seg_id")), int(matcher.group("mod_id")))
            var = lcn_defs.Var.s0_id_to_var(int(matcher.group("id")) - 1)
            value = lcn_defs.VarValue.from_native(int(matcher.group("value")))
            return [ModStatusVar(addr, var, value)]

        matcher = PckParser.PATTERN_VAR_GENERIC.match(data)
        if matcher:
            addr = LcnAddr(int(matcher.group("seg_id")), int(matcher.group("mod_id")))
            var = lcn_defs.Var.UNKNOWN
            value = lcn_defs.VarValue.from_native(int(matcher.group("value")))
            return [ModStatusVar(addr, var, value)]

        matcher = PckParser.PATTERN_THRS5.match(data)
        if matcher:
            ret: list[Input] = []
            addr = LcnAddr(int(matcher.group("seg_id")), int(matcher.group("mod_id")))
            for thrs_id in range(5):
                var = lcn_defs.Var.thrs_id_to_var(0, thrs_id)
                value = lcn_defs.VarValue.from_native(
                    int(matcher.group(f"value{thrs_id + 1}"))
                )
                ret.append(ModStatusVar(addr, var, value))
            return ret

        return None


class ModStatusLedsAndLogicOps(ModInput):
    """Status of LEDs and logic-operations received from an LCN module.

    :param    int      physicalSourceAddr:   The physical source address
    :param    states_led:         The 12 LED states
    :type     states_led:         list(:class:`~pypck.lcn_defs.LedStatus`)
    :param    states_logic_ops:   The 4 logic-operation states
    :type     states_logic_ops:   list(:class:`~pypck.lcn_defs.LogicOpStatus`)
    """

    def __init__(
        self,
        physical_source_addr: LcnAddr,
        states_led: list[lcn_defs.LedStatus],
        states_logic_ops: list[lcn_defs.LogicOpStatus],
    ):
        """Construct ModInput object."""
        super().__init__(physical_source_addr)
        self.states_led = states_led  # 12x LED status.
        self.states_logic_ops = states_logic_ops  # 4x logic-operation status.

    def get_led_state(self, led_id: int) -> lcn_defs.LedStatus:
        """Get the status of a single LED.

        :param    int    led_id:   LED id (0..11)
        :return:                   The LED's status
        :rtype:   list(:class:`~pypck.lcn_defs.LedStatus`)
        """
        return self.states_led[led_id]

    def get_logic_op_state(self, logic_op_id: int) -> lcn_defs.LogicOpStatus:
        """Get the status of a single logic operation.

        :param    int    logic_op_id:    Logic operation id (0..3)
        :return:    The logic-operation's status
        :rtype:     list(:class:`~pypck.lcn_defs.LogicOpStatus`)
        """
        return self.states_logic_ops[logic_op_id]

    @staticmethod
    def try_parse(data: str) -> list[Input] | None:
        """Try to parse the given input text.

        Will return a list of parsed Inputs. The list might be empty (but not
        null).

        :param    data    str:    The input data received from LCN-PCHK

        :return:            The parsed Inputs (never null)
        :rtype:             List with instances of :class:`~pypck.input.Input`
        """
        matcher = PckParser.PATTERN_STATUS_LEDSANDLOGICOPS.match(data)
        if matcher:
            addr = LcnAddr(int(matcher.group("seg_id")), int(matcher.group("mod_id")))

            led_states = matcher.group("led_states").upper()
            states_leds = [lcn_defs.LedStatus(led_state) for led_state in led_states]

            logic_op_states = matcher.group("logic_op_states").upper()
            states_logic_ops = [
                lcn_defs.LogicOpStatus(logic_op_state)
                for logic_op_state in logic_op_states
            ]
            return [ModStatusLedsAndLogicOps(addr, states_leds, states_logic_ops)]

        return None


class ModStatusKeyLocks(ModInput):
    """Status of locked keys received from an LCN module.

    :param    int                physicalSourceAddr:   The source address
    :param    list(list(bool))   states:               The 4x8 key-lock states
    """

    def __init__(self, physical_source_id: LcnAddr, states: list[list[bool]]):
        """Construct ModInput object."""
        super().__init__(physical_source_id)
        self.states = states

    def get_state(self, table_id: int, key_id: int) -> bool:
        """Get the lock-state of a single key.

        :param    int    tableId:    Table id: (0..3  =>  A..D)
        :param    int    keyId:      Key id (0..7  =>  1..8)
        :return:  The key's lock-state
        :rtype:   bool
        """
        return self.states[table_id][key_id]

    @staticmethod
    def try_parse(data: str) -> list[Input] | None:
        """Try to parse the given input text.

        Will return a list of parsed Inputs. The list might be empty (but not
        null).

        :param    data    str:    The input data received from LCN-PCHK

        :return:            The parsed Inputs (never null)
        :rtype:             List with instances of :class:`~pypck.input.Input`
        """
        matcher = PckParser.PATTERN_STATUS_KEYLOCKS.match(data)
        states: list[list[bool]] = []
        if matcher:
            addr = LcnAddr(int(matcher.group("seg_id")), int(matcher.group("mod_id")))
            for i in range(4):
                state = matcher.group(f"table{i:d}")
                if state is not None:
                    states.append(PckParser.get_boolean_value(int(state)))
            return [ModStatusKeyLocks(addr, states)]

        return None


class ModStatusAccessControl(ModInput):
    """Status of a transmitter, transponder or fingerprint sensor."""

    def __init__(
        self,
        physical_source_addr: LcnAddr,
        periphery: lcn_defs.AccessControlPeriphery,
        code: str,
        level: int | None = None,
        key: int | None = None,
        action: lcn_defs.KeyAction | None = None,
        battery: lcn_defs.BatteryStatus | None = None,
    ):
        """Construct ModInput object."""
        super().__init__(physical_source_addr)
        self.periphery = periphery
        self.code = code
        self.level = level
        self.key = key
        self.action = action
        self.battery = battery

    @staticmethod
    def try_parse(data: str) -> list[Input] | None:
        """Try to parse the given input text.

        Will return a list of parsed Inputs. The list might be empty (but not
        null).

        :param    data    str:    The input data received from LCN-PCHK

        :return:            The parsed Inputs (never null)
        :rtype:             List with instances of :class:`~pypck.input.Input`
        """
        matcher = PckParser.PATTERN_STATUS_TRANSMITTER.match(data)
        if matcher:
            periphery = lcn_defs.AccessControlPeriphery.TRANSMITTER
            addr = LcnAddr(int(matcher.group("seg_id")), int(matcher.group("mod_id")))
            code = (
                f"{int(matcher.group('code1')):02x}"
                f"{int(matcher.group('code2')):02x}"
                f"{int(matcher.group('code3')):02x}"
            )
            level = int(matcher.group("level"))
            key = int(matcher.group("key")) - 1

            actions = {
                "1": lcn_defs.KeyAction.HIT,
                "2": lcn_defs.KeyAction.MAKE,
                "3": lcn_defs.KeyAction.BREAK,
            }

            battery_status = {
                "0": lcn_defs.BatteryStatus.FULL,
                "1": lcn_defs.BatteryStatus.WEAK,
            }

            action = actions[matcher.group("action")[2]]
            battery = battery_status[matcher.group("action")[1]]

            return [
                ModStatusAccessControl(
                    addr, periphery, code, level, key, action, battery
                )
            ]

        matcher = PckParser.PATTERN_STATUS_TRANSPONDER.match(data)
        if matcher:
            periphery = lcn_defs.AccessControlPeriphery.TRANSPONDER
            addr = LcnAddr(int(matcher.group("seg_id")), int(matcher.group("mod_id")))
            code = (
                f"{int(matcher.group('code1')):02x}"
                f"{int(matcher.group('code2')):02x}"
                f"{int(matcher.group('code3')):02x}"
            )
            return [ModStatusAccessControl(addr, periphery, code)]

        matcher = PckParser.PATTERN_STATUS_FINGERPRINT.match(data)
        if matcher:
            periphery = lcn_defs.AccessControlPeriphery.FINGERPRINT
            addr = LcnAddr(int(matcher.group("seg_id")), int(matcher.group("mod_id")))
            code = (
                f"{int(matcher.group('code1')):02x}"
                f"{int(matcher.group('code2')):02x}"
                f"{int(matcher.group('code3')):02x}"
            )
            return [ModStatusAccessControl(addr, periphery, code)]

        matcher = PckParser.PATTERN_STATUS_CODELOCK.match(data)
        if matcher:
            periphery = lcn_defs.AccessControlPeriphery.CODELOCK
            addr = LcnAddr(int(matcher.group("seg_id")), int(matcher.group("mod_id")))
            code = (
                f"{int(matcher.group('code1')):02x}"
                f"{int(matcher.group('code2')):02x}"
                f"{int(matcher.group('code3')):02x}"
            )
            return [ModStatusAccessControl(addr, periphery, code)]

        return None


class ModStatusSceneOutputs(ModInput):
    """Status of the output values and ramp values received from an LCN module."""

    def __init__(
        self,
        physical_source_addr: LcnAddr,
        scene_id: int,
        values: list[int],
        ramps: list[int],
    ):
        """Construct ModInput object."""
        super().__init__(physical_source_addr)
        self.scene_id = scene_id
        self.values = values
        self.ramps = ramps

    @staticmethod
    def try_parse(data: str) -> list[Input] | None:
        """Try to parse the given input text.

        Will return a list of parsed Inputs. The list might be empty (but not
        null).

        :param    data    str:    The input data received from LCN-PCHK

        :return:            The parsed Inputs (never null)
        :rtype:             List with instances of :class:`~pypck.input.Input`
        """
        matcher = PckParser.PATTERN_STATUS_SCENE_OUTPUTS.match(data)
        if matcher:
            addr = LcnAddr(int(matcher.group("seg_id")), int(matcher.group("mod_id")))
            scene_id = int(matcher.group("scene_id"))
            values = [int(matcher.group(f"output{i + 1:d}")) for i in range(4)]
            ramps = [int(matcher.group(f"ramp{i + 1:d}")) for i in range(4)]
            return [ModStatusSceneOutputs(addr, scene_id, values, ramps)]

        return None


class ModStatusMotorPositionBS4(ModInput):
    """Status of motor positions (if BS4 connected) received from an LCN module.

    Position and limit is in percent. 0%: cover closed, 100%: cover open.
    """

    def __init__(
        self,
        physical_source_addr: LcnAddr,
        motor: int,
        position: float,
        limit: float | None = None,
        time_down: int | None = None,
        time_up: int | None = None,
    ):
        """Construct ModInput object."""
        super().__init__(physical_source_addr)
        self.motor = motor
        self.position = position
        self.limit = limit
        self.time_down = time_down
        self.time_up = time_up

    @staticmethod
    def try_parse(data: str) -> list[Input] | None:
        """Try to parse the given input text.

        Will return a list of parsed Inputs. The list might be empty (but not
        null).

        :param    data    str:    The input data received from LCN-PCHK

        :return:            The parsed Inputs (never null)
        :rtype:             List with instances of :class:`~pypck.input.Input`
        """
        matcher = PckParser.PATTERN_STATUS_MOTOR_POSITION_BS4.match(data)
        if matcher:
            motor_status_inputs: list[Input] = []
            addr = LcnAddr(int(matcher.group("seg_id")), int(matcher.group("mod_id")))
            for idx in (1, 2):
                motor = matcher.group(f"motor{idx}_id")
                position = matcher.group(f"position{idx}")
                limit = matcher.group(f"limit{idx}")
                time_down = matcher.group(f"time_down{idx}")
                time_up = matcher.group(f"time_up{idx}")

                motor_status_inputs.append(
                    ModStatusMotorPositionBS4(
                        addr,
                        int(motor) - 1,
                        (200 - int(position)) / 2,
                        None if limit == "?" else (200 - int(limit)) / 2,
                        None if time_down == "?" else int(time_down),
                        None if time_up == "?" else int(time_up),
                    )
                )
            return motor_status_inputs

        return None


class ModStatusMotorPositionModule(ModInput):
    """Status of motor positions received from an LCN module.

    Position is in percent. 0%: cover closed, 100%: cover open.
    """

    def __init__(
        self,
        physical_source_addr: LcnAddr,
        motor: int,
        position: float,
    ):
        """Construct ModInput object."""
        super().__init__(physical_source_addr)
        self.motor = motor
        self.position = position

    @staticmethod
    def try_parse(data: str) -> list[Input] | None:
        """Try to parse the given input text.

        Will return a list of parsed Inputs. The list might be empty (but not
        null).

        :param    data    str:    The input data received from LCN-PCHK

        :return:            The parsed Inputs (never null)
        :rtype:             List with instances of :class:`~pypck.input.Input`
        """
        matcher = PckParser.PATTERN_STATUS_MOTOR_POSITION_MODULE.match(data)
        if matcher:
            addr = LcnAddr(int(matcher.group("seg_id")), int(matcher.group("mod_id")))
            motor = matcher.group("motor_id")
            position = matcher.group("position")

            return [
                ModStatusMotorPositionModule(
                    addr,
                    int(motor) - 1,
                    100 - float(position),
                )
            ]

        return None


class ModSendCommandHost(ModInput):
    """Send command to host message from module."""

    def __init__(self, physical_source_addr: LcnAddr, parameters: tuple[int, ...]):
        """Construct ModSendCommandHost object."""
        super().__init__(physical_source_addr)
        self.parameters = parameters

    def get_parameters(self) -> tuple[int, ...]:
        """Return the received parameters.

        :return:    Parameters
        :rtype:     List with parameters of type int.
        """
        return self.parameters

    @staticmethod
    def try_parse(data: str) -> list[Input] | None:
        """Try to parse the given input text.

        Will return a list of parsed Inputs. The list might be empty (but not
        null).

        :param    data    str:    The input data received from LCN-PCHK

        :return:            The parsed Inputs (never null)
        :rtype:             List with instances of :class:`~pypck.input.Input`
        """
        matcher = PckParser.PATTERN_SEND_COMMAND_HOST.match(data)
        if matcher:
            addr = LcnAddr(int(matcher.group("seg_id")), int(matcher.group("mod_id")))
            parameters = tuple(
                int(param) for param in matcher.groups()[3:] if param is not None
            )
            return [ModSendCommandHost(addr, parameters)]

        return None


class ModSendKeysHost(ModInput):
    """Send command to host message from module."""

    def __init__(
        self,
        physical_source_addr: LcnAddr,
        actions: list[lcn_defs.SendKeyCommand],
        keys: list[bool],
    ):
        """Construct ModSendKeysHost object."""
        super().__init__(physical_source_addr)
        self.actions = actions
        self.keys = keys

    def get_actions(self) -> list[lcn_defs.SendKeyCommand]:
        """Get key actions for each table.

        :returns:   List of length 3 with key actions for each table A, B, C.
        :rtype:     list(:class:`~pypck.lcn_defs.SendKeyCommand`)
        """
        return self.actions

    def get_keys(self) -> list[bool]:
        """Get keys which should be triggered.

        :returns:   List of booleans (length 8) for each key
                    (True: trigger, False: do nothing).
        :rtype:     list(bool)
        """
        return self.keys

    @staticmethod
    def try_parse(data: str) -> list[Input] | None:
        """Try to parse the given input text.

        Will return a list of parsed Inputs. The list might be empty (but not
        null).

        :param    data    str:    The input data received from LCN-PCHK

        :return:            The parsed Inputs (never null)
        :rtype:             List with instances of :class:`~pypck.input.Input`
        """
        matcher = PckParser.PATTERN_SEND_KEYS_HOST.match(data)
        if matcher:
            addr = LcnAddr(int(matcher.group("seg_id")), int(matcher.group("mod_id")))
            actions_value = int(matcher.group("actions"))
            keys_value = int(matcher.group("keys"))

            mapping = (
                lcn_defs.SendKeyCommand.DONTSEND,
                lcn_defs.SendKeyCommand.HIT,
                lcn_defs.SendKeyCommand.MAKE,
                lcn_defs.SendKeyCommand.BREAK,
            )

            actions = []
            for idx in range(3):
                action = mapping[(actions_value >> 2 * idx) & 0x03]
                actions.append(action)

            keys = [bool(keys_value >> bit & 0x01) for bit in range(8)]

            return [ModSendKeysHost(addr, actions, keys)]

        return None


# ## Other inputs


class Unknown(Input):
    """Handle all unknown input data."""

    def __init__(self, data: str):
        """Construct Input object."""
        super().__init__()
        self._data = data

    @staticmethod
    def try_parse(data: str) -> list[Input]:
        """Try to parse the given input text.

        Will return a list of parsed Inputs. The list might be empty (but not
        null).

        :param    data    str:    The input data received from LCN-PCHK

        :return:            The parsed Inputs (never null)
        :rtype:             List with instances of :class:`~pypck.input.Input`
        """
        return [Unknown(data)]

    @property
    def data(self) -> str:
        """Return the received data.

        :return:    Received data.
        :rtype:     str
        """
        return self._data


class InputParser:
    """Parse all input objects for given input data."""

    parsers: list[type[Input]] = [
        AuthUsername,
        AuthPassword,
        AuthOk,
        AuthFailed,
        LcnConnState,
        LicenseError,
        DecModeSet,
        CommandError,
        Ping,
        ModAck,
        ModNameComment,
        ModSk,
        ModSn,
        ModStatusGroups,
        ModStatusOutput,
        ModStatusOutputNative,
        ModStatusRelays,
        ModStatusBinSensors,
        ModStatusVar,
        ModStatusLedsAndLogicOps,
        ModStatusKeyLocks,
        ModStatusAccessControl,
        ModStatusSceneOutputs,
        ModStatusMotorPositionBS4,
        ModStatusMotorPositionModule,
        ModSendCommandHost,
        ModSendKeysHost,
        Unknown,
    ]

    @staticmethod
    def parse(data: str) -> list[Input]:
        """Parse all input objects for given input data.

        :param    str    data:    The input data received from LCN-PCHK

        :return:    The parsed Inputs (never null)
        :rtype:     List with instances of :class:`~pypck.input.Input`
        """
        for parser in InputParser.parsers:
            returns: list[Input] | None = parser.try_parse(data)
            if returns is not None:
                for ret in returns:
                    ret.pck = data
                return returns

        # We must never get to this point since the Unknown parser matches
        # everything.
        assert False



"""Test the status requester of a module connection."""

import asyncio
from unittest.mock import call

from pypck import inputs
from pypck.pck_commands import PckGenerator
from pypck.status_requester import StatusRequest

from .conftest import MockDeviceConnection

RELAY_STATES = [True, False, True, False, True, False, True, False]


async def test_request_status(module10: MockDeviceConnection) -> None:
    """Test requesting the status of a module."""
    request_task = asyncio.create_task(
        module10.status_requester.request(
            response_type=inputs.ModStatusRelays,
            request_pck=PckGenerator.request_relays_status(),
            max_age=0,
        )
    )
    await asyncio.sleep(0)

    module10.send_command.assert_awaited_with(
        False, PckGenerator.request_relays_status()
    )

    await module10.async_process_input(
        inputs.ModStatusRelays(module10.addr, RELAY_STATES)
    )

    result = await request_task

    assert isinstance(result, inputs.ModStatusRelays)
    assert result.physical_source_addr == module10.addr
    assert result.states == RELAY_STATES


async def test_request_status_stored(module10: MockDeviceConnection) -> None:
    """Test requesting the status of a module with stored status request."""
    status_request = StatusRequest(
        type=inputs.ModStatusRelays,
        parameters=frozenset(),
        timestamp=asyncio.get_running_loop().time(),
        response=asyncio.get_running_loop().create_future(),
    )
    status_request.response.set_result(
        inputs.ModStatusRelays(module10.addr, RELAY_STATES)
    )
    module10.status_requester.request_cache.add(status_request)

    result = await module10.status_requester.request(
        response_type=inputs.ModStatusRelays,
        request_pck=PckGenerator.request_relays_status(),
        max_age=10,
    )

    assert isinstance(result, inputs.ModStatusRelays)
    assert result.physical_source_addr == module10.addr
    assert result.states == RELAY_STATES
    assert (
        call(False, PckGenerator.request_relays_status())
        not in module10.send_command.await_args_list
    )


async def test_request_status_expired(module10: MockDeviceConnection) -> None:
    """Test requesting the status of a module with stored status request but max_age expired."""
    states = [False] * 8
    status_request = StatusRequest(
        type=inputs.ModStatusRelays,
        parameters=frozenset(),
        timestamp=asyncio.get_running_loop().time() - 10,
        response=asyncio.get_running_loop().create_future(),
    )
    status_request.response.set_result(inputs.ModStatusRelays(module10.addr, states))
    module10.status_requester.request_cache.add(status_request)

    request_task = asyncio.create_task(
        module10.status_requester.request(
            response_type=inputs.ModStatusRelays,
            request_pck=PckGenerator.request_relays_status(),
            max_age=5,
        )
    )
    await asyncio.sleep(0)

    module10.send_command.assert_awaited_with(
        False, PckGenerator.request_relays_status()
    )

    await module10.async_process_input(
        inputs.ModStatusRelays(module10.addr, RELAY_STATES)
    )

    result = await request_task

    assert isinstance(result, inputs.ModStatusRelays)
    assert result.physical_source_addr == module10.addr
    assert result.states == RELAY_STATES



"""Connection classes for pypck."""

from __future__ import annotations

import asyncio
import contextlib
import logging
from collections.abc import Callable, Iterable
from types import TracebackType
from typing import Any

from pypck import inputs, lcn_defs
from pypck.device import DeviceConnection
from pypck.helpers import TaskRegistry
from pypck.lcn_addr import LcnAddr
from pypck.lcn_defs import LcnEvent
from pypck.pck_commands import PckGenerator

_LOGGER = logging.getLogger(__name__)


class PchkLicenseError(Exception):
    """Exception which is raised if a license error occurred."""

    def __init__(self, message: str | None = None):
        """Initialize instance."""
        if message is None:
            message = (
                "License Error: Maximum number of connections was reached. An "
                "additional license key is required."
            )
        super().__init__(message)


class PchkAuthenticationError(Exception):
    """Exception which is raised if authentication failed."""

    def __init__(self, message: str | None = None):
        """Initialize instance."""
        if message is None:
            message = "Authentication failed"
        super().__init__(message)


class PchkConnectionRefusedError(Exception):
    """Exception which is raised if connection was refused."""

    def __init__(self, message: str | None = None):
        """Initialize instance."""
        if message is None:
            message = "Connection refused"
        super().__init__(message)


class PchkConnectionFailedError(Exception):
    """Exception which is raised if connection was refused."""

    def __init__(self, message: str | None = None):
        """Initialize instance."""
        if message is None:
            message = "Connection failed"
        super().__init__(message)


class PchkLcnNotConnectedError(Exception):
    """Exception which is raised if there is no connection to the LCN bus."""

    def __init__(self, message: str | None = None):
        """Initialize instance."""
        if message is None:
            message = "LCN not connected."
        super().__init__(message)


class PchkConnectionManager:
    """Connection to LCN-PCHK."""

    last_ping: float
    ping_timeout_handle: asyncio.TimerHandle | None
    authentication_completed_future: asyncio.Future[bool]
    license_error_future: asyncio.Future[bool]

    def __init__(
        self,
        host: str,
        port: int,
        username: str,
        password: str,
        settings: dict[str, Any] | None = None,
        connection_id: str = "PCHK",
    ) -> None:
        """Construct PchkConnectionManager."""
        self.task_registry = TaskRegistry()
        self.host = host
        self.port = port
        self.connection_id = connection_id

        self.reader: asyncio.StreamReader | None = None
        self.writer: asyncio.StreamWriter | None = None
        self.buffer: asyncio.Queue[bytes] = asyncio.Queue()
        self.last_bus_activity = asyncio.get_running_loop().time()

        self.username = username
        self.password = password

        # Settings
        if settings is None:
            settings = {}
        self.settings = lcn_defs.default_connection_settings
        self.settings.update(settings)

        self.idle_time = self.settings["BUS_IDLE_TIME"]
        self.ping_send_delay = self.settings["PING_SEND_DELAY"]
        self.ping_recv_timeout = self.settings["PING_RECV_TIMEOUT"]
        self.ping_timeout_handle = None
        self.ping_counter = 0
        self.dim_mode = self.settings["DIM_MODE"]
        self.status_mode = lcn_defs.OutputPortStatusMode.PERCENT

        self.is_lcn_connected = True
        self.local_seg_id = 0

        # Events, Futures, Locks for synchronization
        self.segment_scan_completed_event = asyncio.Event()
        self.authentication_completed_future = asyncio.Future()
        self.license_error_future = asyncio.Future()
        self.module_serial_number_received = asyncio.Lock()
        self.segment_coupler_response_received = asyncio.Lock()

        # All modules from or to a communication occurs are represented by a
        # unique ModuleConnection object.  All ModuleConnection objects are
        # stored in this dictionary.  Communication to groups is handled by
        # GroupConnection object that are created on the fly and not stored
        # permanently.
        self.device_connections: dict[LcnAddr, DeviceConnection] = {}
        self.segment_coupler_ids: list[int] = []

        self.input_callbacks: set[Callable[[inputs.Input], None]] = set()
        self.event_callbacks: set[Callable[[LcnEvent], None]] = set()
        self.register_for_events(self.event_callback)

    # Socket read/write

    async def read_data_loop(self) -> None:
        """Process incoming data."""
        assert self.reader is not None
        assert self.writer is not None
        loop = asyncio.get_running_loop()
        _LOGGER.debug("Read data loop started")
        try:
            while not self.writer.is_closing():
                try:
                    data = await self.reader.readuntil(
                        PckGenerator.TERMINATION.encode()
                    )
                    self.last_bus_activity = loop.time()
                except (
                    asyncio.IncompleteReadError,
                    TimeoutError,
                    OSError,
                ):
                    _LOGGER.debug("Connection to %s lost", self.connection_id)
                    self.fire_event(LcnEvent.CONNECTION_LOST)
                    await self.async_close()
                    break

                try:
                    message = data.decode("utf-8").split(PckGenerator.TERMINATION)[0]
                except UnicodeDecodeError as err:
                    try:
                        message = data.decode("cp1250").split(PckGenerator.TERMINATION)[
                            0
                        ]
                        _LOGGER.warning(
                            "Incorrect PCK encoding detected, possibly caused by LinHK: %s - PCK recovered using cp1250",
                            err,
                        )
                    except UnicodeDecodeError as err2:
                        _LOGGER.warning(
                            "PCK decoding error: %s - skipping received PCK message",
                            err2,
                        )
                        continue
                await self.process_message(message)
        finally:
            _LOGGER.debug("Read data loop closed")

    async def write_data_loop(self) -> None:
        """Process queue and write data."""
        assert self.writer is not None
        loop = asyncio.get_running_loop()
        try:
            _LOGGER.debug("Write data loop started")
            while not self.writer.is_closing():
                data = await self.buffer.get()
                while (loop.time() - self.last_bus_activity) < self.idle_time:
                    await asyncio.sleep(self.idle_time)

                _LOGGER.debug(
                    "to %s: %s",
                    self.connection_id,
                    data.decode().rstrip(PckGenerator.TERMINATION),
                )
                self.writer.write(data)
                await self.writer.drain()
                self.last_bus_activity = loop.time()
        finally:
            # empty the queue
            while not self.buffer.empty():
                await self.buffer.get()
            _LOGGER.debug("Write data loop closed")

    # Open/close connection, authentication & setup.

    async def async_connect(self, timeout: float = 30) -> None:
        """Establish a connection to PCHK at the given socket."""
        self.authentication_completed_future = asyncio.Future()
        self.license_error_future = asyncio.Future()

        _LOGGER.debug(
            "Starting connection attempt to %s server at %s:%d",
            self.connection_id,
            self.host,
            self.port,
        )

        done: Iterable[asyncio.Future[Any]]
        pending: Iterable[asyncio.Future[Any]]
        done, pending = await asyncio.wait(
            (
                asyncio.create_task(self.open_connection()),
                self.license_error_future,
                self.authentication_completed_future,
            ),
            timeout=timeout,
            return_when=asyncio.FIRST_EXCEPTION,
        )

        # Raise any exception which occurs
        # (ConnectionRefusedError, PchkAuthenticationError, PchkLicenseError)
        for awaitable in done:
            if not awaitable.cancelled():
                if exc := awaitable.exception():
                    await self.async_close()
                    if isinstance(exc, (ConnectionRefusedError, OSError)):
                        raise PchkConnectionRefusedError()
                    else:
                        raise exc

        if pending:
            for awaitable in pending:
                awaitable.cancel()
            await self.async_close()
            raise PchkConnectionFailedError()

        if not self.is_lcn_connected:
            raise PchkLcnNotConnectedError()

        # start segment scan
        await self.scan_segment_couplers(
            self.settings["SK_NUM_TRIES"], self.settings["DEFAULT_TIMEOUT"]
        )

    async def open_connection(self) -> None:
        """Connect to PCHK server (no authentication or license error check)."""
        self.reader, self.writer = await asyncio.open_connection(self.host, self.port)

        address = self.writer.get_extra_info("peername")
        _LOGGER.debug("%s server connected at %s:%d", self.connection_id, *address)

        # main write loop
        self.task_registry.create_task(self.write_data_loop())

        # main read loop
        self.task_registry.create_task(self.read_data_loop())

    async def async_close(self) -> None:
        """Close the active connection."""
        if self.ping_timeout_handle is not None:
            self.ping_timeout_handle.cancel()
        await self.task_registry.cancel_all_tasks()
        if self.writer:
            self.writer.close()
            with contextlib.suppress(OSError):  # occurs when TCP connection is lost
                await self.writer.wait_closed()

        _LOGGER.debug("Connection to %s closed.", self.connection_id)

    async def wait_closed(self) -> None:
        """Wait until connection to PCHK server is closed."""
        if self.writer is not None:
            await self.writer.wait_closed()

    async def __aenter__(self) -> PchkConnectionManager:
        """Context manager enter method."""
        await self.async_connect()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        exc_traceback: TracebackType | None,
    ) -> None:
        """Context manager exit method."""
        await self.async_close()

    async def on_auth(self, success: bool) -> None:
        """Is called after successful authentication."""
        if success:
            _LOGGER.debug("%s authorization successful!", self.connection_id)
            self.authentication_completed_future.set_result(True)
            # Try to set the PCHK decimal mode
            await self.send_command(PckGenerator.set_dec_mode(), to_host=True)
        else:
            _LOGGER.debug("%s authorization failed!", self.connection_id)
            self.authentication_completed_future.set_exception(PchkAuthenticationError)

    async def on_license_error(self) -> None:
        """Is called if a license error occurs during connection."""
        _LOGGER.debug("%s: License Error.", self.connection_id)
        self.license_error_future.set_exception(PchkLicenseError())

    async def on_successful_login(self) -> None:
        """Is called after connection to LCN bus system is established."""
        _LOGGER.debug("%s login successful.", self.connection_id)
        await self.send_command(
            PckGenerator.set_operation_mode(self.dim_mode, self.status_mode),
            to_host=True,
        )
        self.task_registry.create_task(self.ping())

    async def lcn_connection_status_changed(self, is_lcn_connected: bool) -> None:
        """Set the current connection state to the LCN bus."""
        self.is_lcn_connected = is_lcn_connected
        self.fire_event(LcnEvent.BUS_CONNECTION_STATUS_CHANGED)
        if is_lcn_connected:
            _LOGGER.debug("%s: LCN is connected.", self.connection_id)
            self.fire_event(LcnEvent.BUS_CONNECTED)
        else:
            _LOGGER.debug("%s: LCN is not connected.", self.connection_id)
            self.fire_event(LcnEvent.BUS_DISCONNECTED)

    async def ping_received(self, count: int | None) -> None:
        """Ping was received."""
        if self.ping_timeout_handle is not None:
            self.ping_timeout_handle.cancel()
        self.last_ping = asyncio.get_running_loop().time()

    def is_ready(self) -> bool:
        """Retrieve the overall connection state."""
        return self.segment_scan_completed_event.is_set()

    # Addresses, modules and groups

    def set_local_seg_id(self, local_seg_id: int) -> None:
        """Set the local segment id."""
        old_local_seg_id = self.local_seg_id

        self.local_seg_id = local_seg_id
        # replace all device_connections with current local_seg_id with new
        # local_seg_id
        for addr in list(self.device_connections):
            if addr.seg_id == old_local_seg_id:
                address_conn = self.device_connections.pop(addr)
                address_conn.addr = LcnAddr(
                    self.local_seg_id, addr.addr_id, addr.is_group
                )
                self.device_connections[address_conn.addr] = address_conn

    def physical_to_logical(self, addr: LcnAddr) -> LcnAddr:
        """Convert the physical segment id of an address to the logical one."""
        return LcnAddr(
            self.local_seg_id if addr.seg_id in (0, 4) else addr.seg_id,
            addr.addr_id,
            addr.is_group,
        )

    def get_device_connection(self, addr: LcnAddr) -> DeviceConnection:
        """Create and/or return a connection to the given module or group."""
        if addr.seg_id == 0 and self.local_seg_id != -1:
            addr = LcnAddr(self.local_seg_id, addr.addr_id, addr.is_group)

        device_connection = self.device_connections.get(addr, None)
        if device_connection is None:
            device_connection = DeviceConnection(
                self,
                addr,
                wants_ack=False if addr.is_group else self.settings["ACKNOWLEDGE"],
            )
            self.device_connections[addr] = device_connection

        return device_connection

    # Other

    async def dump_modules(self) -> dict[str, dict[str, dict[str, Any]]]:
        """Dump all modules and information about them in a JSON serializable dict."""
        dump: dict[str, dict[str, dict[str, Any]]] = {}
        for address_conn in self.device_connections.values():
            seg = f"{address_conn.addr.seg_id:d}"
            addr = f"{address_conn.addr.addr_id}"
            if seg not in dump:
                dump[seg] = {}
            dump[seg][addr] = await address_conn.dump_details()
        return dump

    # Command sending / retrieval.

    async def send_command(
        self, pck: bytes | str, to_host: bool = False, **kwargs: Any
    ) -> bool:
        """Send a PCK command to the PCHK server."""
        if not self.is_lcn_connected and not to_host:
            return False

        assert self.writer is not None
        if not self.writer.is_closing():
            if isinstance(pck, str):
                data = (pck + PckGenerator.TERMINATION).encode()
            else:
                data = pck + PckGenerator.TERMINATION.encode()
            await self.buffer.put(data)
            return True
        return False

    async def process_message(self, message: str) -> None:
        """Is called when a new text message is received from the PCHK server."""
        _LOGGER.debug("from %s: %s", self.connection_id, message)
        inps = inputs.InputParser.parse(message)

        if inps is not None:
            for inp in inps:
                await self.async_process_input(inp)

    async def async_process_input(self, inp: inputs.Input) -> None:
        """Process an input command."""
        # Inputs from Host
        if isinstance(inp, inputs.AuthUsername):
            await self.send_command(self.username, to_host=True)
        elif isinstance(inp, inputs.AuthPassword):
            await self.send_command(self.password, to_host=True)
        elif isinstance(inp, inputs.AuthOk):
            await self.on_auth(True)
        elif isinstance(inp, inputs.AuthFailed):
            await self.on_auth(False)
        elif isinstance(inp, inputs.LcnConnState):
            await self.lcn_connection_status_changed(inp.is_lcn_connected)
        elif isinstance(inp, inputs.LicenseError):
            await self.on_license_error()
        elif isinstance(inp, inputs.DecModeSet):
            self.license_error_future.set_result(True)
            await self.on_successful_login()
        elif isinstance(inp, inputs.CommandError):
            _LOGGER.debug("LCN command error: %s", inp.message)
        elif isinstance(inp, inputs.Ping):
            await self.ping_received(inp.count)
        elif isinstance(inp, inputs.ModSk):
            if inp.physical_source_addr.seg_id == 0:
                self.set_local_seg_id(inp.reported_seg_id)
            if self.segment_coupler_response_received.locked():
                self.segment_coupler_response_received.release()
            # store reported segment coupler id
            if inp.reported_seg_id not in self.segment_coupler_ids:
                self.segment_coupler_ids.append(inp.reported_seg_id)
        elif isinstance(inp, inputs.Unknown):
            return

        # Inputs from bus
        elif self.is_ready():
            if isinstance(inp, inputs.ModInput):
                logical_source_addr = self.physical_to_logical(inp.physical_source_addr)
                if not logical_source_addr.is_group:
                    module_conn = self.get_device_connection(logical_source_addr)
                    if isinstance(inp, inputs.ModSn):
                        # used to extend scan_modules() timeout
                        if self.module_serial_number_received.locked():
                            self.module_serial_number_received.release()

                    await module_conn.async_process_input(inp)

            # Forward all known inputs to callback listeners.
            for input_callback in self.input_callbacks:
                input_callback(inp)

    async def ping(self) -> None:
        """Send pings."""
        assert self.writer is not None
        while not self.writer.is_closing():
            await self.send_command(f"^ping{self.ping_counter:d}", to_host=True)
            self.ping_timeout_handle = asyncio.get_running_loop().call_later(
                self.ping_recv_timeout, lambda: self.fire_event(LcnEvent.PING_TIMEOUT)
            )
            self.ping_counter += 1
            await asyncio.sleep(self.ping_send_delay)

    async def scan_modules(self, num_tries: int = 3, timeout: float = 3) -> None:
        """Scan for modules on the bus.

        This is a convenience coroutine which handles all the logic when
        scanning modules on the bus. Because of heavy bus traffic, not all
        modules might respond to a scan command immediately.
        The coroutine will make 'num_tries' attempts to send a scan command
        and waits 'timeout' after the last module response before
        proceeding to the next try.
        """
        segment_coupler_ids = (
            self.segment_coupler_ids if self.segment_coupler_ids else [0]
        )

        for _ in range(num_tries):
            for segment_id in segment_coupler_ids:
                if segment_id == self.local_seg_id:
                    segment_id = 0
                await self.send_command(
                    PckGenerator.generate_address_header(
                        LcnAddr(segment_id, 3, True), self.local_seg_id, True
                    )
                    + PckGenerator.empty()
                )

            # Wait loop which is extended on every serial number received
            while True:
                try:
                    await asyncio.wait_for(
                        self.module_serial_number_received.acquire(),
                        timeout,
                    )
                except TimeoutError:
                    break

    async def scan_segment_couplers(
        self, num_tries: int = 3, timeout: float = 1.5
    ) -> None:
        """Scan for segment couplers on the bus.

        This is a convenience coroutine which handles all the logic when
        scanning segment couplers on the bus. Because of heavy bus traffic,
        not all segment couplers might respond to a scan command immediately.
        The coroutine will make 'num_tries' attempts to send a scan command
        and waits 'timeout' after the last segment coupler response
        before proceeding to the next try.
        """
        for _ in range(num_tries):
            await self.send_command(
                PckGenerator.generate_address_header(
                    LcnAddr(3, 3, True), self.local_seg_id, False
                )
                + PckGenerator.segment_coupler_scan()
            )

            # Wait loop which is extended on every segment coupler response
            while True:
                try:
                    await asyncio.wait_for(
                        self.segment_coupler_response_received.acquire(),
                        timeout,
                    )
                except TimeoutError:
                    break

        # No segment coupler expected (num_tries=0)
        if len(self.segment_coupler_ids) == 0:
            _LOGGER.debug("%s: No segment coupler found.", self.connection_id)

        self.segment_scan_completed_event.set()

    # Callbacks for inputs and events

    def register_for_inputs(
        self, callback: Callable[[inputs.Input], None]
    ) -> Callable[..., None]:
        """Register a function for callback on PCK message received.

        Returns a function to unregister the callback.
        """
        self.input_callbacks.add(callback)
        return lambda callback=callback: self.input_callbacks.remove(callback)

    def fire_event(self, event: LcnEvent) -> None:
        """Fire event."""
        for event_callback in self.event_callbacks:
            event_callback(event)

    def register_for_events(
        self, callback: Callable[[lcn_defs.LcnEvent], None]
    ) -> Callable[..., None]:
        """Register a function for callback on LCN events.

        Return a function to unregister the callback.
        """
        self.event_callbacks.add(callback)
        return lambda callback=callback: self.event_callbacks.remove(callback)

    def event_callback(self, event: LcnEvent) -> None:
        """Handle events from PchkConnection."""
        _LOGGER.debug("%s: LCN-Event: %s", self.connection_id, event)
